"""Spawn subprocess asynchronously."""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
import asyncio
from asyncio import subprocess
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
import locale
import logging
from pathlib import Path
import shlex

from pypyr.config import config
from pypyr.errors import ContextError, MultiError
from pypyr.subproc import SimpleCommandTypes, SubprocessResult

logger = logging.getLogger(__name__)

# region windows shlex
is_windows = config.is_windows

DEFAULT_ENCODING = (config.default_cmd_encoding
                    if config.default_cmd_encoding
                    else locale.getpreferredencoding(False))

# this code is in fact covered when run on windows (during CI)
# set no cover so no complaining from coverage on posix
if is_windows:  # pragma: no cover
    from ctypes.wintypes import LPWSTR, HLOCAL
    from ctypes import windll, c_int, POINTER, byref  # type: ignore

    CommandLineToArgvW = windll.shell32.CommandLineToArgvW
    CommandLineToArgvW.restype = POINTER(LPWSTR)

    LocalFree = windll.kernel32.LocalFree
    LocalFree.argtypes = [HLOCAL]
    LocalFree.restype = HLOCAL


def winshlex_split(cmd: str) -> list[str]:  # pragma: no cover
    """Split cmd string into args using Windows api."""
    num_args = c_int()
    lp_args = CommandLineToArgvW(cmd, byref(num_args))

    # copy response
    args = [lp_args[i] for i in range(num_args.value)]

    # must free allocated block of contiguous memory, prevent leaks
    if LocalFree(lp_args):
        # return value is None when LocalFree succeeds
        # on fail return value is handle to the local memory object
        raise RuntimeError(
            f"Failed to release {lp_args} after winddll.shell32 "
            + "CommandLineToArgvW")

    return args

# endregion windows shlex


shlexer = winshlex_split if is_windows else shlex.split


class Command:
    """A subprocess run instruction. Use async run() to spawn the subprocess.

    Attributes:
        cmd (str/bytes/Pathlike or list[str/bytes/Pathlike]): A str for a
            single run instruction, or a list of str for >1 run instructions.
        is_shell (bool): Invoke subprocess through the default shell.
        cwd (Path-like): Current working directory to this during execution.
        is_save (bool): Save output to self.results.
        is_text (bool): Treat stdin, stdout & stderr streams as text in the
            provided encoding, not bytes. This is only relevant when
            is_save is True.
        stdout (None | str | Pathlike): Write stdout here. None means inherit
            the parent's handle, '/dev/null' means to write to the null device,
            and a path will write to file.
        stderr (None | str | Pathlike): Write stderr here. None means inherit
            the parent's handle, '/dev/null' means to write to the null device,
            '/dev/stdout' redirects to stdout, and a path will write to file.
        encoding (str): Use this encoding on all the output streams. Default to
            the system's default encoding. Only applicable if is_save True.
        append (bool): If stdout/stderr refers to a file path, append to file
            rather than overwrite if it exists. Only relevant when is_save
            False.
        results (list[SubprocessResult]): List of results. Populated with the
            result of each run instruction in `cmd`. Only when is_save is True.
    """

    def __init__(self,
                 cmd,
                 is_shell=False,
                 cwd=None,
                 is_save=False,
                 is_text=True,
                 stdout=None,
                 stderr=None,
                 encoding=None,
                 append=False):
        """Initialize the Cmd."""
        self.cmd = cmd
        self.is_shell = is_shell
        self.cwd = cwd
        self.is_save = is_save
        self.is_text = is_text if is_save else False

        if is_save:
            if stdout or stderr:
                raise ContextError(
                    "You can't set `stdout` or `stderr` when `save` is True.")
            self.stdout = self.stderr = subprocess.PIPE
        else:
            self.stdout = stdout
            self.stderr = stderr
        self.encoding = encoding if encoding else DEFAULT_ENCODING
        self.append = append

        self._results: list[SubprocessResult | Exception | list] = []

    @property
    def results(self) -> list[SubprocessResult | Exception | list]:
        """Results of this Command after run()."""
        return self._results

    @contextmanager
    def output_handles(self):
        """Return stdout + stderr output handles for subprocess.

        If self.stdout or self.stderr are None, will return None for whichever
        is None. This will result in the subprocess using the parent's handler
        for that output device.

        If self.stdout or self.stderr refers to a path, will open the file(s)
        at the given location for writing. Will create parent directories if
        these do not exist. self.append controls whether to overwrite or append
        files. Set self.encoding to write in text mode in the given encoding,
        otherwise the file handle opens in binary mode.

        The context manager will close the file handles (if any) when it exits.

        If self.stdout or self.stderr are the special values subprocess.PIPE,
        '/dev/null' or '/dev/stdout', will return the appropriate subprocess
        constants for these.

        Returns:
            (stdout, stderr): stdout & stderr handles to use when executing
                this command as a subprocess.
        """
        stdout = self.stdout
        stderr = self.stderr
        # fast short-circuit when no processing necessary
        if not stdout and not stderr:
            #  this works because stdout/stderr will be PIPE when capturing
            logger.debug("stdout & stderr attaching to parent process.")
            yield stdout, stderr
        else:
            # append or overwrite
            mode = 'ab' if self.append else 'wb'
            files = []
            try:
                # stdout
                if stdout == subprocess.PIPE:
                    logger.debug("capturing stdout")
                elif stdout == '/dev/null':
                    logger.debug("redirecting stdout to /dev/null")
                    stdout = subprocess.DEVNULL
                elif stdout:
                    # by elimination, it's a path
                    logger.debug("writing stdout to %s in mode '%s'",
                                 stdout, mode)
                    Path(stdout).parent.mkdir(parents=True, exist_ok=True)
                    stdout = open(stdout, mode=mode)
                    files.append(stdout)

                #  stderr
                if stderr == subprocess.PIPE:
                    logger.debug("capturing stderr")
                elif stderr == '/dev/null':
                    logger.debug("redirecting stderr to /dev/null")
                    stderr = subprocess.DEVNULL
                elif stderr == '/dev/stdout':
                    logger.debug("redirecting stderr to /dev/stdout")
                    stderr = subprocess.STDOUT
                elif stderr:
                    # by elimination, it's a path
                    logger.debug("writing stderr to %s in mode '%s'",
                                 stderr, mode)
                    Path(stderr).parent.mkdir(parents=True, exist_ok=True)
                    stderr = open(stderr, mode=mode)
                    files.append(stderr)

                yield stdout, stderr
            finally:
                for f in files:
                    logger.debug("closing cmd file output %s", f.name)
                    f.close()

    def parse_results(self) -> list[Exception]:
        """Parse subprocess output for errors.

        Errors are likely to be SubprocessError where returncode != 0.

        If there were problems getting to the subprocess itself (i.e it
        couldn't even run), likely to be a PermissionError or OSError.

        This does not raise any exceptions, it just returns a list of them.

        Returns:
            list[Exception]: Flattened list of Exception objects.
        """
        errors = []
        for result in self._results:
            for error in self._parse_result(result):
                errors.append(error)
        return errors

    def _parse_result(self,
                      result: SubprocessResult | Exception | list
                      ) -> Iterator[Exception]:
        """Parse results recursively looking for errors."""
        if isinstance(result, Exception):
            yield result
        elif isinstance(result, SubprocessResult):
            error = result.check_returncode()
            if error:
                yield error
        elif isinstance(result, list):
            for nested_result in result:
                yield from self._parse_result(nested_result)
        else:
            # this really should be unreachable.
            raise TypeError(
                f"{result} is {type(result)}.\n"
                + "It should be SubprocessResult | Exception | "
                + "list[SubprocessResult | Exception].")
        # don't just add stuff here! remember those yields further up!

    async def run(self) -> None:
        """Run the command asynchronously as a subprocess.

        Do NOT raise exceptions in here. Add exceptions to self.results
        instead.

        Typical exceptions that end up in `results`:
            pypyr.errors.ContextError: The cmd executable instruction is
                formatted incorrectly.
            subprocess.CalledProcessError: Error executing the subprocess.
            FileNotFoundError: The system could not find the executable. It
                might not be on $PATH, or the path to the executable might be
                wrong.
        """
        cmd = self.cmd
        try:
            # this here because all outputs for this cmd write to the same
            # device/file handles - whether single cmd or list of commands.
            with self.output_handles() as (stdout, stderr):
                if isinstance(cmd, SimpleCommandTypes):
                    # just one command, run as is
                    try:
                        result = await self._run(cmd, stdout, stderr)
                    except Exception as ex:
                        # mypy currently doesn't like re-using local var name
                        result = ex  # type: ignore

                    self._results.append(result)
                elif isinstance(cmd, Sequence):
                    # Sequence won't be str - explicitly checked before in if
                    # except weirdness if passing bytearray (ByteString), but
                    # why would you?
                    tasks = [self._run(c, stdout, stderr) for c in cmd]
                    results = await asyncio.gather(*tasks,
                                                   return_exceptions=True)
                    # extend results shared state here coz concurrent now done
                    self._results.extend(results)
                else:
                    err = ContextError(
                        f"{cmd} cmds input.\n"
                        "Each command should be a simple string, or a "
                        "sub-list to run in serial:\n"
                        "cmds:\n"
                        "  - ./my-executable --arg\n"
                        "  - ./another-executable --arg\n"
                        "  - [./executableA.1, ./executableA.2]\n\n"
                        "Or in the expanded syntax, set `run` to a list:\n"
                        "cmds:\n"
                        "  run:\n"
                        "    - ./my-executable --arg\n"
                        "    - ./another-executable --arg\n"
                        "    - [./executableA.1, ./executableA.2]\n"
                        "  cwd: ./mydir")
                    self._results.append(err)
        except Exception as ex:
            # instead of raising exception, any errs MUST be available on
            # `results` for this particular cmd (otherwise caller gather of >1
            # commands won't know _which_ cmd raised the err.)
            self._results.append(ex)

    async def _run(self, cmd,
                   stdout=None,
                   stderr=None
                   ) -> SubprocessResult | list[SubprocessResult | Exception]:
        """Spawn subprocess for cmd with stdout + stderr output streams.

        If self.is_save is True, will append the completed process results to
        self.results, for both success and failure. Will strip trailing
        whitespace from output.

        A non-zero returncode will not raise error, it returns a
                SubprocessResult with the details instead. Will raise error
                if subprocess couldn't even execute at all.

        Args:
            cmd (str | Pathlike | list): The executable + args to spawn as
                                a subprocess. If list, will execute cmds in
                                list in serial.
            stdout: Write stdout here. This can be None (attach to parent
                    process), an open file handle, subprocess.PIPE (to capture)
                    or subprocess.DEVNULL.
            stderr: Write stderr here. This can be None (attach to parent
                    process), an open file handle, or subprocess.PIPE (to
                    capture) or subprocess.DEVNULL, or subprocess.STDOUT to
                    redirect to stdout.

        Returns:
            A SubprocessResult or list of SubprocessResult when cmd is a list
            of serial commands.

        Raises:
            subprocess.CalledProcessError: Error executing the subprocess.
            FileNotFoundError: The system could not find the executable. It
                might not be on $PATH, or the path to the executable might be
                wrong.
        """
        if isinstance(cmd, (list, tuple)):
            logger.debug(
                "async %s is a series of commands that will run in serial",
                cmd)
            results: list[SubprocessResult | Exception] = []
            try:
                for serial_cmd in cmd:
                    result = await self._spawn(serial_cmd,
                                               stdout=stdout,
                                               stderr=stderr)
                    results.append(result)
                    # don't proceed to next serial cmd if return code != 0
                    if result.returncode:
                        break
            except Exception as ex:
                # need to save results as they happen, so if later in sequence
                # fails, can still return results of preceding serial cmds
                results.append(ex)
            return results

        return await self._spawn(cmd, stdout=stdout, stderr=stderr)

    async def _spawn(self, cmd, stdout, stderr) -> SubprocessResult:
        if self.cwd:
            logger.debug("Processing command string in dir %s: %s",
                         self.cwd, self.cmd)
        else:
            logger.debug("Processing command string: %s", cmd)

        # errs from _inside_ the subprocess will go to stderr and raise
        # via check_returncode. errs finding the executable will raise
        # right here.
        if self.is_shell:
            proc = await asyncio.create_subprocess_shell(cmd,
                                                         stdout=stdout,
                                                         stderr=stderr,
                                                         cwd=self.cwd)
        else:
            cmd = shlexer(cmd)  # type: ignore
            logger.debug("arg split is: %s", cmd)
            proc = await asyncio.create_subprocess_exec(cmd[0],
                                                        *cmd[1:],
                                                        stdout=stdout,
                                                        stderr=stderr,
                                                        cwd=self.cwd)

        stdout_data, stderr_data = await proc.communicate()

        if self.is_save:
            if self.is_text:
                if stdout_data:
                    stdout_data = stdout_data.decode(
                        self.encoding).rstrip()  # type: ignore

                if stderr_data:
                    stderr_data = stderr_data.decode(
                        self.encoding).rstrip()  # type: ignore

        return SubprocessResult(cmd=cmd, returncode=proc.returncode,
                                stdout=stdout_data, stderr=stderr_data)

    def __eq__(self, other):
        """Check equality for all attributes."""
        if self is other:
            return True

        if type(other) is Command:
            return self.__dict__ == other.__dict__

        return NotImplemented


class Commands():
    """Execute a bunch of Command objects asynchronously.

    Use .append to add a Command to run.

    run() runs every Command appended with .append() in parallel.

    Check results with the `results` property.
    """

    def __init__(self) -> None:
        """Initialize commands to run asynchronously."""
        self.commands: list[Command] = []
        self._results: list[SubprocessResult | Exception | list] = []  # None
        self._is_save: bool = False

    def __getitem__(self, index: int) -> Command:
        """Get Command in collection by index."""
        return self.commands[index]

    def __iter__(self) -> Iterator[Command]:
        """Allow iteration through Command instances contained in Commands."""
        return iter(self.commands)

    def __len__(self) -> int:
        """Length (count) of Command instances in Commands."""
        return len(self.commands)

    def append(self, cmd: Command) -> None:
        """Append a Command to run asynchronously."""
        if cmd.is_save and not self._is_save:
            self._is_save = True
        self.commands.append(cmd)

    @property
    def is_save(self) -> bool:
        """Is true when any of the Commands to process has is_save==True."""
        return self._is_save

    @property
    def results(self) -> list[SubprocessResult | Exception
                              | list[SubprocessResult | Exception]]:
        """Return the aggregate list of results for all the command runs.

        Only contains results where a Command.is_save == True.
        """
        return self._results

    def run(self) -> None:
        """Run all commands as asynchronous subprocesses.

        This is the method that does the work. It contains an asyncio.run(),
        so you can't call this from an already existing event-loop.

        When this is done, whether it raises an exception or not, you can check
        `results` on the instance to see outputs for each command.

        Raises:
            pypyr.errors.MultiError: Aggregate error containing a list of
                any/all errors that any of the subprocesses might have raised.
        """
        asyncio.run(self._run())
        errors = []
        for cmd in self.commands:
            if cmd.is_save:
                self._results.extend(cmd._results)
            errs = cmd.parse_results()
            if errs:
                errors.extend(errs)

        if errors:
            raise MultiError(("The following error(s) occurred while running "
                              + "the async commands:"),
                             errors)

    async def _run(self) -> None:
        """Run all commands asynchronously & gather results."""
        tasks = [cmd.run() for cmd in self.commands]
        await asyncio.gather(*tasks)

    def __eq__(self, other):
        """Check equality for Command contained in commands list."""
        if self is other:
            return True

        if type(other) is Commands:
            return self.commands == other.commands

        return NotImplemented



"""py.py unit tests."""
import pytest
from pypyr.context import Context
from pypyr.dsl import PyString
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
from pypyr.moduleloader import _ChainMapPretendDict
import pypyr.steps.py

# region py


def test_py_existing_key():
    """Py expression can update existing key."""
    context = Context({'x': 123, 'py': "x = abs(-42); save('x')"})
    assert context['x'] == 123
    pypyr.steps.py.run_step(context)
    # nothing added to context, but x is updated.
    assert context == {'x': 42, 'py': "x = abs(-42); save('x')"}


def test_py_with_import():
    """Py expression can use imports."""
    context = Context(
        {'y': 4, 'py': 'import math\nx = math.sqrt(y)\nsave("x")'})
    assert context['y'] == 4
    assert 'x' not in context
    pypyr.steps.py.run_step(context)
    assert context['x'] == 2
    assert context['y'] == 4


def test_py_single_code():
    """One word python function works."""
    context = Context({'py': 'abs(-1-2)'})
    pypyr.steps.py.run_step(context)


def test_py_sequence():
    """Sequence of py code works and touches context."""
    context = Context({'py': "test = 1; save('test')"})
    pypyr.steps.py.run_step(context)

    context.update({'py': "test += 2; save('test')"})
    pypyr.steps.py.run_step(context)

    context.update({'py': "test += 3; save('test')"})
    pypyr.steps.py.run_step(context)

    assert context['test'] == 6, "context should be 6 at this point"

    assert context == {'py': "test += 3; save('test')", 'test': 6}


def test_py_replace_objects():
    """Rebind mutable and immutable objects in exec scope."""
    pycode = """\
alist = [0, 1, 2]
adict = {'a': 'b', 'c': 'd'}
anint = 456
mutate_me.append(12)
astring = 'updated'
"""
    context = Context({'alist': [0, 1],
                       'adict': {'a': 'b'},
                       'anint': 123,
                       'mutate_me': [10, 11],
                       'astring': 'a string',
                       'py': pycode})

    pypyr.steps.py.run_step(context)

    # rebound vars not changed because no `save()`
    assert context == {'alist': [0, 1],
                       'adict': {'a': 'b'},
                       'anint': 123,
                       'mutate_me': [10, 11, 12],
                       'astring': 'a string',
                       'py': pycode}


def test_py_with_closure_scope():
    """Free variables scopes resolve to global."""
    pycode = """\
list1.append(1)
list2 = ['a', 'b', 'c']
out = [(x, y) for x in list1 for y in list2]
save('out')
"""
    context = Context({'list1': [0],
                       'list2': ['a', 'b'],
                       'py': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'list1': [0, 1],
                       'list2': ['a', 'b'],
                       'py': pycode,
                       'out': [(0, 'a'), (0, 'b'), (0, 'c'),
                               (1, 'a'), (1, 'b'), (1, 'c')]}


def test_py_save_inputs():
    """Special save takes args and kwargs."""
    pycode = """\
a = 1
b = 2
c = 3
d = 4
e = 5
f = 6
g = 7
h = 8
i = 9
j = 10
save('a')
save('b', 'c')
save(dd=d)
save(ee=e, ff=f+1)
save('g', 'h', ii=i, jj=j, k=5)
"""
    context = Context({'py': pycode})
    pypyr.steps.py.run_step(context)

    assert context == {'a': 1,
                       'b': 2,
                       'c': 3,
                       'dd': 4,
                       'ee': 5,
                       'ff': 7,
                       'g': 7,
                       'h': 8,
                       'ii': 9,
                       'jj': 10,
                       'k': 5,
                       'py': pycode}


def test_py_save_existing():
    """Save overwrites existing keys."""
    pycode = """\
alist = [0, 1, 2]
adict = {'a': 'b', 'c': 'd'}
anint = 456
mutate_me.append(12)
astring = 'updated'

save('alist', 'adict', 'anint', 'mutate_me', 'astring')
"""
    context = Context({'alist': [0, 1],
                       'adict': {'a': 'b'},
                       'anint': 123,
                       # strictly speaking not necessary to save mutate_me
                       # explicitly, coz it's mutated not rebound, but test
                       # that it works coz end-users mightn't see the
                       # distinction
                       'mutate_me': [10, 11],
                       'astring': 'a string',
                       'py': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'alist': [0, 1, 2],
                       'adict': {'a': 'b', 'c': 'd'},
                       'anint': 456,
                       'mutate_me': [10, 11, 12],
                       'astring': 'updated',
                       'py': pycode}


def test_py_save_key_error():
    """Save raises special message on KeyError."""
    pycode = "notfound='arb'; save(notfound)"
    context = Context({'py': pycode})

    with pytest.raises(KeyError) as err:
        pypyr.steps.py.run_step(context)

    # KeyError looks like it reprs on the __str__?
    expected = repr(
        "Trying to save 'arb', but can't find it in the py step scope. "
        "Remember it should be save('key'), not save(key) - mind the quotes.")

    assert str(err.value) == expected


def test_py_save_keyword_exists():
    """Save does not overwrite existing user 'save' key."""
    pycode = "k2='v2'; save('k2')"
    context = Context({'save': 123, 'py': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'save': 123, 'py': pycode, 'k2': 'v2'}


def test_py_imports_not_in_context():
    """Imported modules + non saves don't end up in context."""
    pycode = """\
from math import sqrt
b = sqrt(a)
"""
    context = Context({'a': 4,
                       'py': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'a': 4,
                       'py': pycode}


def test_py_sequence_with_semicolons():
    """Single py code string with semi - colons works."""
    context = Context({'py':
                       'x = abs(-1); x+= abs(-2); x += abs(-3); save("x")'})
    pypyr.steps.py.run_step(context)

    assert context == Context({
        'py': 'x = abs(-1); x+= abs(-2); x += abs(-3); save("x")',
        'x': 6})


def test_py_sequence_with_linefeeds():
    """Single py code string with linefeeds works."""
    context = Context({'py':
                       'abs(-1)\nabs(-2)\nabs(-3)'})
    pypyr.steps.py.run_step(context)


def test_py_error_throws():
    """Input pycode error should raise up to caller."""
    with pytest.raises(AssertionError):
        context = Context({'py': 'assert False'})
        pypyr.steps.py.run_step(context)


def test_py_no_context_throw():
    """No pycode in context should throw assert error."""
    with pytest.raises(KeyNotInContextError) as err_info:
        context = Context({'blah': 'blah blah'})
        pypyr.steps.py.run_step(context)

    assert str(err_info.value) == ("context['py'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.py.")


def test_py_none_context_throw():
    """None pycode in context should throw assert error."""
    with pytest.raises(KeyInContextHasNoValueError):
        context = Context({'py': None})
        pypyr.steps.py.run_step(context)
# endregion py

# region pycode


def test_pycode_single_code():
    """One word python function works."""
    context = Context({'pycode': 'abs(-1-2)'})
    pypyr.steps.py.run_step(context)


def test_pycode_sequence():
    """Sequence of py code works and touches context."""
    context = Context({'pycode': "context['test'] = 1;"})
    pypyr.steps.py.run_step(context)

    context.update({'pycode': "context['test'] += 2"})
    pypyr.steps.py.run_step(context)

    context.update({'pycode': "context['test'] += 3"})
    pypyr.steps.py.run_step(context)

    assert context['test'] == 6, "context should be 6 at this point"

    assert context == {'pycode': "context['test'] += 3", 'test': 6}


def test_pycode_replace_objects():
    """Rebind mutable and immutable objects in eval scope."""
    pycode = """\
context['alist'] = [0, 1, 2]
context['adict'] = {'a': 'b', 'c': 'd'}
context['anint'] = 456
context['mutate_me'].append(12)
context['astring'] = 'updated'
"""
    context = Context({'alist': [0, 1],
                       'adict': {'a': 'b'},
                       'anint': 123,
                       'mutate_me': [10, 11],
                       'astring': 'a string',
                       'pycode': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'alist': [0, 1, 2],
                       'adict': {'a': 'b', 'c': 'd'},
                       'anint': 456,
                       'mutate_me': [10, 11, 12],
                       'astring': 'updated',
                       'pycode': pycode}


def test_pycode_with_closure_scope():
    """Free variables in closure scopes resolve to global."""
    pycode = """\
context['list1'].append(1)
context['list2'] = ['a', 'b', 'c']
context['out'] = [(x, y) for x in context['list1'] for y in context['list2']]
"""
    context = Context({'list1': [0],
                       'list2': ['a', 'b'],
                       'pycode': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'list1': [0, 1],
                       'list2': ['a', 'b', 'c'],
                       'pycode': pycode,
                       'out': [(0, 'a'), (0, 'b'), (0, 'c'),
                               (1, 'a'), (1, 'b'), (1, 'c')]}


def test_pycode_imports_not_in_context():
    """Imported modules don't end up in context."""
    pycode = """\
from math import sqrt
context['b'] = sqrt(context['a'])
"""
    context = Context({'a': 4,
                       'pycode': pycode})

    pypyr.steps.py.run_step(context)

    assert context == {'a': 4,
                       'b': 2,
                       'pycode': pycode}


def test_pycode_sequence_with_semicolons():
    """Single py code string with semi - colons works."""
    context = Context({'pycode':
                       'abs(-1); abs(-2); abs(-3);'})
    pypyr.steps.py.run_step(context)

    assert context == {'pycode':
                       'abs(-1); abs(-2); abs(-3);'}, ("context in and out "
                                                       "the same")


def test_pycode_sequence_with_linefeeds():
    """Single py code string with linefeeds works."""
    context = Context({'pycode':
                       'abs(-1)\nabs(-2)\nabs(-3)'})
    pypyr.steps.py.run_step(context)


def test_pycode_error_throws():
    """Input pycode error should raise up to caller."""
    with pytest.raises(AssertionError):
        context = Context({'pycode': 'assert False'})
        pypyr.steps.py.run_step(context)


def test_pycode_no__context_throw():
    """No pycode in context should throw assert error."""
    with pytest.raises(KeyNotInContextError) as err_info:
        context = Context({'blah': 'blah blah'})
        pypyr.steps.py.run_step(context)

    assert str(err_info.value) == ("context['py'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.py.")


def test_pycode_none_context_throw():
    """None pycode in context should throw assert error."""
    with pytest.raises(KeyInContextHasNoValueError):
        context = Context({'pycode': None})
        pypyr.steps.py.run_step(context)

# endregion pycode

# region nested scope for both py and pycode


def test_py_scope_class_accessors():
    """Class with instance, static & class attributes for py and pycode."""
    # region py
    pycode = """\
class ArbClassForEvalTest():
    a = 123

    def __init__(self):
        self.b = 456

    def dothing(self, val):
        return ArbClassForEvalTest.a + val

    @classmethod
    def dothing_class_method(cls, val):
        return cls.a - val

    @staticmethod
    def dothing_static_method(val):
        return val + 2

out_class = ArbClassForEvalTest
a = ArbClassForEvalTest.a
b = ArbClassForEvalTest().b
c = ArbClassForEvalTest().dothing(2)
d = ArbClassForEvalTest.dothing_class_method(3)
e = ArbClassForEvalTest.dothing_static_method(4)
f = ArbClassForEvalTest

save('a', 'b', 'c', 'd', 'e', 'f')
"""
    context = Context({'a': 999,
                       'py': pycode})
    pypyr.steps.py.run_step(context)

    def verify_context(context):
        assert context['a'] == 123
        assert context['b'] == 456
        assert context['c'] == 125
        assert context['d'] == 120
        assert context['e'] == 6

        instance = context['f']()
        assert instance.b == 456
        assert instance.dothing(5) == 128

        assert len(context) == 7

    verify_context(context)
    assert PyString('f().dothing(6)').get_value(context) == 129

    # endregion py
    # region pycode
    pycode = """\
class ArbClassForEvalTest():
    a = 123

    def __init__(self):
        self.b = 456

    def dothing(self, val):
        return ArbClassForEvalTest.a + val

    @classmethod
    def dothing_class_method(cls, val):
        return cls.a - val

    @staticmethod
    def dothing_static_method(val):
        return val + 2

out_class = ArbClassForEvalTest
context['a'] = ArbClassForEvalTest.a
context['b'] = ArbClassForEvalTest().b
context['c'] = ArbClassForEvalTest().dothing(2)
context['d'] = ArbClassForEvalTest.dothing_class_method(3)
context['e'] = ArbClassForEvalTest.dothing_static_method(4)
context['f'] = ArbClassForEvalTest
"""
    context = Context({'a': 999,
                       'pycode': pycode})
    pypyr.steps.py.run_step(context)

    verify_context(context)
    assert PyString('f.dothing_class_method(6)').get_value(context) == 117
    # endregion pycode


def test_py_scope_function():
    """Nested functions resolve globals."""
    # region py
    pycode = """\
a = context_in_a + 2

def my_function(arg1):
    return arg1 + context_in_b

def my_closure(arg1):
    closure_scoped = a + arg1 + context_in_c

    def enclosed(arg2):
        return arg1 + arg2 + closure_scoped + context_in_d

    return enclosed

save('a', 'my_function', closure=my_closure(3))
"""
    context = Context({'context_in_a': 10,
                       'context_in_b': 20,
                       'context_in_c': 30,
                       'context_in_d': 40,
                       'py': pycode})
    pypyr.steps.py.run_step(context)

    def verify_context(context):
        assert context['a'] == 12
        assert context['my_function'](3) == 23
        # closure_scoped == 45
        assert context['closure'](4) == 92

        assert len(context) == 8

    verify_context(context)
    assert PyString('closure(5) - abs(-3)').get_value(context) == 90

    # endregion py

    # region pycode
    pycode = """\
a = context['context_in_a'] + 2

def my_function(arg1):
    return arg1 + context['context_in_b']

def my_closure(arg1):
    closure_scoped = a + arg1 + context['context_in_c']

    def enclosed(arg2):
        return arg1 + arg2 + closure_scoped + context['context_in_d']

    return enclosed

context['a'] = a
context['my_function'] = my_function
context['closure'] = my_closure(3)
"""
    context = Context({'context_in_a': 10,
                       'context_in_b': 20,
                       'context_in_c': 30,
                       'context_in_d': 40,
                       'pycode': pycode})
    pypyr.steps.py.run_step(context)

    verify_context(context)
    assert PyString('closure(5) - abs(-3)').get_value(context) == 90
    # endregion pycode


def test_py_scope_class_implicit_global():
    """Class-level global reference resolves."""
    # region py
    # will resolve because globals is a dict
    pycode = """\
class MyClass():
    b = context_in_a + 1

    def dothing(self, val):
        return val + context_in_a

    @classmethod
    def dothing_class_method(cls, val):
        return val + context_in_a

    @staticmethod
    def dothing_static_method(val):
        return val + context_in_a

assert MyClass.b == 11
assert MyClass().dothing(10) == 20
assert MyClass.dothing_class_method(20) == 30
assert MyClass.dothing_static_method(30) == 40
save('MyClass')
"""
    context = Context({'context_in_a': 10,
                       'py': pycode})

    pypyr.steps.py.run_step(context)

    assert len(context) == 3
    assert context['context_in_a'] == 10
    assert context['py'] == pycode
    assert 'MyClass' in context

    context['context_in_a'] = 100
    assert context['MyClass']().dothing(40) == 50
    assert context['MyClass'].dothing_class_method(50) == 60
    assert context['MyClass'].dothing_static_method(60) == 70
    assert context['context_in_a'] == 100
    # endregion py

    # region pycode
    # Does resolve for pycode form, because it uses a standard dict as globals
    pycode = """\
context['bare'] = 111
class MyClass():
    b = context['context_in_a'] + 1
    context['context_in_b'] = abs(-22)

    def dothing(self, val):
        context['dothing_instance'] = val + 2

    @classmethod
    def dothing_class_method(cls, val):
        context['dothing_class'] = val + 3

    @staticmethod
    def dothing_static_method(val):
        context['dothing_static'] = val + 4

assert MyClass.b == 11
MyClass().dothing(10)
MyClass.dothing_class_method(11)
MyClass.dothing_static_method(12)
"""
    context = Context({'context_in_a': 10,
                       'context_in_b': 20,
                       'pycode': pycode})
    pypyr.steps.py.run_step(context)

    assert context == {'bare': 111,
                       'context_in_a': 10,
                       'context_in_b': 22,
                       'dothing_instance': 12,
                       'dothing_class': 14,
                       'dothing_static': 16,
                       'pycode': pycode}
    # endregion pycode


def test_py_scope_class_globals():
    """Global lookup inside class."""
    # region py
    pycode = """\
a = context_in_a + 2
context_in_b = 22

class MyClass():
    b = 3
    global context_in_c
    c = context_in_c + abs(-3)
    context_in_c = 333

    def my_function(self, arg1):
        return arg1 + context_in_c + context_in_d

    class MyNested():
        global context_in_e
        d = context_in_e + 5
        context_in_e = 555

        def dothing(self):
            return context_in_a + 6

assert a == 12
# notice context_in_c is updated at class level
assert MyClass().my_function(1) == 333 + context_in_d + 1
assert MyClass.b == 3
assert MyClass.c == 33
assert MyClass.MyNested.d == 55
assert MyClass.MyNested().dothing() == 16
assert context_in_c == 333
assert context_in_e == 555
context_in_e = 5555
save('context_in_e')
"""
    context = Context({'context_in_a': 10,
                       'context_in_b': 20,
                       'context_in_c': 30,
                       'context_in_d': 40,
                       'context_in_e': 50,
                       'py': pycode})
    pypyr.steps.py.run_step(context)

    assert context == {'context_in_a': 10,
                       'context_in_b': 20,  # no save, no update
                       'context_in_c': 30,  # in class scope
                       'context_in_d': 40,
                       'context_in_e': 5555,
                       'py': pycode}
    # endregion py

    # region py style with _ChainMapPretendDict to illustrate difference
    pycode = """\
a = context_in_a + 2
context_in_b = 22

class MyClass():
    b = 3
    global context_in_c
    c = context_in_c + abs(-3)
    context_in_c = 333

    def my_function(self, arg1):
        return arg1 + context_in_c + context_in_d

    class MyNested():
        global context_in_e
        d = context_in_e + 5
        context_in_e = 555

        def dothing(self):
            return context_in_a + 6

assert a == 12
# notice context_in_c still thinks it's 30 - global doesn't update
assert MyClass().my_function(1) == 30 + context_in_d + 1
assert MyClass.b == 3
assert MyClass.c == 33
assert MyClass.MyNested.d == 55
assert MyClass.MyNested().dothing() == 16
assert context_in_c == 30
assert context_in_e == 50 # even with global, no update coz ChainMap
context_in_e = 5555
# save('context_in_e')
"""
    context = Context({'context_in_a': 10,
                       'context_in_b': 20,
                       'context_in_c': 30,
                       'context_in_d': 40,
                       'context_in_e': 50,
                       'py': pycode})
    first_dict = {}
    # first_dict gets any adds/imports, not the 2nd dict which is context.
    # other than this, first_dict behaves like locals, basically.
    globals = _ChainMapPretendDict(first_dict,
                                   context)
    # the save function ref allows pipeline to use save to persist vars
    # back to context,
    # first_dict['save'] = get_save(context, globals)
    exec(context['py'], globals)

    assert context == {'context_in_a': 10,
                       'context_in_b': 20,  # no save, no update
                       'context_in_c': 30,  # in class scope
                       'context_in_d': 40,
                       'context_in_e': 50,  # even with global, no update
                       'py': pycode}
    # endregion  py style with _ChainMapPretendDict to illustrate difference

    # region pycode
    pycode = """\
a = context['context_in_a'] + 2
context['context_in_b'] = 22

class MyClass():
    b = 3
    # no global necessary for pycode form
    c = context['context_in_c'] + abs(-3)
    context['context_in_c'] = 333

    def my_function(self, arg1):
        return arg1 + context['context_in_c'] + context['context_in_d']

    class MyNested():
        d = context['context_in_e'] + 5
        context['context_in_e'] = 555

        def dothing(self):
            return context['context_in_a'] + 6

assert a == 12
assert MyClass().my_function(1) == 333 + 40 + 1
assert MyClass.b == 3
assert MyClass.c == 33
assert MyClass.MyNested.d == 55
assert MyClass.MyNested().dothing() == 16
assert context['context_in_c'] == 333
context['context_in_e'] = 5555
"""
    context = Context({'context_in_a': 10,
                       'context_in_b': 20,
                       'context_in_c': 30,
                       'context_in_d': 40,
                       'context_in_e': 50,
                       'pycode': pycode})
    pypyr.steps.py.run_step(context)

    assert context == {'context_in_a': 10,
                       'context_in_b': 22,
                       'context_in_c': 333,  # in class scope
                       'context_in_d': 40,
                       'context_in_e': 5555,
                       'pycode': pycode}
    # endregion pycode

# endregion nested scope for both py and pycode



"""namespacecache.py unit tests."""
from sys import intern
from unittest.mock import patch

import pypyr.cache.namespacecache as namespacecache
from pypyr.moduleloader import ImportVisitor

# region parse_and_load


def test_get_namespace_parse_and_load():
    """Parse and load returns namespace dict."""
    source = """\
import math
import tests.arbpack.arbmod
from tests.arbpack import arbmod2
from tests.arbpack import arbmod3 as ab3
from tests.arbpack.arbmultiattr import arb_attr as x, arb_func as y
"""
    out = namespacecache.NamespaceCache().get_namespace(source)

    assert len(out) == 6
    assert all(key in out for key in ['math', 'tests', 'arbmod2', 'ab3',
                                      'x', 'y'])


# endregion parse_and_load

# region NamespaceCache

# region NamespaceCache: get_namespace


def test_get_namespace_cache_hit():
    """Get namespace interns and hits cache."""
    with patch.object(ImportVisitor,
                      'get_namespace',
                      return_value='ns') as mock:
        ns = namespacecache.NamespaceCache().get_namespace("arbkey")

    mock.assert_called_once_with('arbkey')
    assert ns == "ns"


def test_get_namespace_same_obj():
    """Get namespace returns same obj for same source."""
    source = "from tests.arbpack import arbmod3 as ab3"
    cache = namespacecache.NamespaceCache()
    ns = cache.get_namespace(source)
    assert len(ns) == 1
    assert ns['ab3'].arb_func_in_arbmod3(321) == 321

    ns2 = cache.get_namespace("from tests.arbpack import arbmod3 as ab3")
    assert ns2 is ns
    assert len(ns) == 1
    assert ns['ab3'].arb_func_in_arbmod3(321) == 321

    intern("from tests.arbpack import arbmod3 as ab3") is list(
        cache._cache.keys())[0]

# endregion NamespaceCache: get_namespace

# endregion NamespaceCache



"""Unit tests for tests/common/path_mock.py."""
import os
from pathlib import Path
from unittest.mock import call, patch

from tests.common.path_mock import FakePath, MultiPath


def test_fake_path():
    """A FakePath behaves like a Path."""
    fp = FakePath('/arb')
    assert isinstance(fp, Path)
    assert isinstance(fp, os.PathLike)
    assert fp.__fspath__() == f'{os.sep}arb'
    assert str(fp) == f'{os.sep}arb'

    # mocked out methods don't touch the file system
    assert fp.exists() is False
    fp.mkdir()
    fp.mkdir(exist_ok=True)
    fp.mkdir(exist_ok=False, parents=True)
    assert fp.is_file() is False
    assert fp.is_dir() is False

    # delegate to underlying path
    assert fp.name == 'arb'
    fp.mock_instance.joinpath.assert_not_called()
    assert fp.joinpath('sub') == Path('/arb/sub')

    # calls recorded
    assert fp.mock.mock_calls == [call('/arb'),
                                  call().exists(),
                                  call().mkdir(),
                                  call().mkdir(exist_ok=True),
                                  call().mkdir(exist_ok=False, parents=True),
                                  call().is_file(),
                                  call().is_dir(),
                                  call().joinpath('sub')]

    # can use various mock assert helpers
    # mock_instance is same as return_value
    fp.mock.assert_called_once_with('/arb')
    fp.mock_instance.mkdir.assert_called_with(exist_ok=False,
                                              parents=True)
    fp.mock.return_value.joinpath.assert_called_once_with('sub')
    fp.mock.return_value.is_dir.assert_called_once()
    fp.mock_instance.is_dir.assert_called_once()


def test_fake_path_parent():
    """Methods returning Paths return FakePaths."""
    fp = FakePath('/arb/sub')
    parent = fp.parent
    assert fp.mock.mock_calls == [call('/arb/sub')]
    assert fp.mock.parent.mock_calls == []
    assert type(parent) is FakePath
    assert parent == Path('/arb')


def test_fake_path_constructor_overrides():
    """A FakePath with special values for is_dir, is_file."""
    # multi constructor
    fp = FakePath('arb', 'sub')
    assert fp == Path('arb/sub')

    f = FakePath('myfile', is_file=True)
    assert f.is_file() is True
    assert f.is_dir() is False
    assert f.exists() is True
    f.mock.assert_called_once_with('myfile')
    f.mock_instance.is_file.assert_called_once()

    d = FakePath('mydir', is_dir=True)
    assert d.is_dir() is True
    assert d.exists() is True
    d.mock.assert_called_once_with('mydir')
    d.mock_instance.is_file.assert_not_called()
    d.mock_instance.is_dir.assert_called_once()
    assert d.is_file() is False


def test_multi_fake_path():
    """Use multi path fakes with patch."""
    og_path_type = Path
    with patch('tests.common.path_mock_test.Path',
               new=MultiPath()) as mock_paths:
        p = Path('/arb')
        assert type(p) is FakePath
        assert isinstance(p, og_path_type)
        assert isinstance(p, Path)
        assert p.is_dir() is False
        assert p.is_file() is False
        assert p.exists() is False
        sp = p.joinpath('sub')
        sp2 = p.joinpath('sub', 'sub2')

        p2 = Path('/arb2')
        assert not p2.exists()

    assert sp == Path('/arb/sub')
    assert sp2 == Path('/arb/sub/sub2')

    assert len(mock_paths.instances) == 2
    assert mock_paths.instances[0].mock.mock_calls == [call('/arb'),
                                                       call().is_dir(),
                                                       call().is_file(),
                                                       call().exists(),
                                                       call().joinpath('sub'),
                                                       call().joinpath('sub',
                                                                       'sub2')]

    mock_paths.instances[1].mock.assert_called_once_with('/arb2')
    mock_paths.instances[1].mock.return_value.exists.assert_called_once()


def test_multi_fake_path_objs():
    """Multiple objects under the MultiPath scope."""
    with patch('tests.common.path_mock_test.Path',
               new=MultiPath()) as mock_paths:
        p = Path('/arb')
        assert not p.is_dir()
        assert not p.exists()
        assert not Path('arb2', 'sub2').is_file()

    assert mock_paths.instances[0].mock.mock_calls == [call('/arb'),
                                                       call().is_dir(),
                                                       call().exists()]

    assert mock_paths.instances[1].mock.mock_calls == [call('arb2', 'sub2'),
                                                       call().is_file()]


def test_multi_fake_path_with_known():
    """Use multi path fakes with patch and pre-configured known paths."""
    multi = MultiPath({'/arb': FakePath('/arb', is_dir=True),
                       ('arb2', 'sub2'): FakePath('/a/s', is_file=True)})

    with patch('tests.common.path_mock_test.Path', new=multi):
        p = Path('arb2', 'sub2')
        assert isinstance(p, Path)
        assert p.is_file()
        assert Path('/arb').is_dir()

    assert p == Path('/a/s')

    mock1 = multi.instances[0].mock
    mock1.assert_called_once_with('/a/s')
    mock1.return_value.is_file.assert_called_once()
    mock1.return_value.is_dir.assert_not_called()

    mock2 = multi.instances[1].mock
    mock2.assert_called_once_with('/arb')
    mock2.return_value.exists.assert_not_called()
    mock2.return_value.is_dir.assert_called_once()



"""contextclearall.py unit tests."""
import builtins

from pypyr.context import Context
import pypyr.steps.contextclearall


def test_context_clear_all_pass():
    """Context Clear All removes everything from context."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'key4': 'value4',
        'contextClear': [
            'key2',
            'key4',
            'contextClear'
        ]
    })

    # len is only the one added item 'a'
    assert context.pystring_globals_update({'a': 'b'}) == 1

    pypyr.steps.contextclearall.run_step(context)

    assert len(context) == 0
    assert context is not None
    assert isinstance(context, Context)

    assert type(context._pystring_globals) is dict
    assert context._pystring_globals == {}
    assert list(dict.items(context._pystring_namespace)) == [
        ('__builtins__', builtins.__dict__)]

    assert context._pystring_namespace.maps == [{}, {}]

    context['k1'] = 'value1'

    assert context['k1'] == 'value1'
    # global namespace still works.
    assert context.get_eval_string('len(k1)') == 6


def test_context_clear_all_context_empty_already():
    """Context Clear All removes everything from context from empty."""
    context = Context()

    pypyr.steps.contextclearall.run_step(context)

    assert len(context) == 0
    assert len(context._pystring_globals) == 0
    assert context._pystring_globals == {}



"""Naive file loader for integration testing."""
from pathlib import Path
import pypyr.yaml

working_dir_tests = Path.cwd().joinpath('tests')


def get_pipeline_definition(pipeline_name, parent):
    """Simplified file loader for integration testing."""
    with open(working_dir_tests.joinpath(
            f'{pipeline_name}.yaml'), encoding='utf-8') as yaml_file:
        return pypyr.yaml.get_pipeline_yaml(yaml_file)



"""dict.py unit tests."""
import pypyr.parser.dict


def test_arg_string_parses_to_argdict():
    """Args input kvp string should return dictionary."""
    out = pypyr.parser.dict.get_parsed_context(['key1=value1',
                                                'key 2=value2',
                                                'key3=value3'])
    arg_dict = out['argDict']
    assert arg_dict
    assert arg_dict['key1'] == 'value1', "key1 should be value1."
    assert arg_dict['key 2'] == 'value2', "key2 should be value2."
    assert arg_dict['key3'] == 'value3', "key3 should be value3."
    assert len(out) == 1
    assert len(arg_dict) == 3, "3 items expected"


def test_arg_string_parses_to_single_entry_argdict():
    """Input kvp string should return dictionary with 1 item."""
    out = pypyr.parser.dict.get_parsed_context(['key 1=value 2 value3'])
    arg_dict = out['argDict']
    assert arg_dict
    assert arg_dict['key 1'] == 'value 2 value3'
    assert len(out) == 1, "1 item expected"
    assert len(arg_dict) == 1, "1 item expected"


def test_no_equals_string_parses_to_single_entry_argdict():
    """No equals input kvp parses to key=''."""
    out = pypyr.parser.dict.get_parsed_context(['key1value2,value3'])
    assert out == {'argDict': {'key1value2,value3': ''}}


def test_argdict_no_equals_combined_with_equals():
    """No equals input kvp string parses to key='' and key=value."""
    out = pypyr.parser.dict.get_parsed_context(
        ['key1value2,value3', 'key 1=value 1'])
    assert out == {'argDict': {'key1value2,value3': '',
                               'key 1': 'value 1'}}


def test_argdict_no_equals_combined_with_double_equals():
    """Ignore subsequent equals."""
    out = pypyr.parser.dict.get_parsed_context(
        ['key1value2,value3', 'key 1=value 1=123'])
    assert out == {'argDict': {'key1value2,value3': '',
                               'key 1': 'value 1=123'}}


def test_empty_string_empty_dict_argdict():
    """Empty input string should return empty argDict."""
    out = pypyr.parser.dict.get_parsed_context(None)

    assert out
    assert not out['argDict']
    assert out['argDict'] == {}



"""pypyr/cache/admin.py unit tests."""
import pypyr.cache.admin as cache_admin

from pypyr.cache.backoffcache import backoff_cache
from pypyr.cache.filecache import file_cache
from pypyr.cache.loadercache import loader_cache
from pypyr.cache.namespacecache import pystring_namespace_cache
from pypyr.cache.parsercache import contextparser_cache
from pypyr.cache.stepcache import step_cache
from pypyr.retries import builtin_backoffs

# region load_backoff_callable


def test_cache_clear_all():
    """Clear all cache."""
    cache_admin.clear_all()

    assert backoff_cache._cache == builtin_backoffs
    assert file_cache._cache == {}
    assert loader_cache._cache == {}
    assert pystring_namespace_cache._cache == {}
    assert contextparser_cache._cache == {}
    assert step_cache._cache == {}

    backoff_cache.get_backoff('tests.arbpack.arbcallables.ArbCallable')
    assert len(backoff_cache._cache) == len(builtin_backoffs) + 1

    # testing full file cache clear in pypyr/loaders/file_test.py in:
    # test_get_pipeline_definition_clear_all_cache
    file_cache._cache['arb'] = 'delete me'
    assert len(file_cache._cache) == 1

    loader_cache.get_pype_loader('tests.arbpack.arbloader')
    assert len(loader_cache._cache) == 1

    pystring_namespace_cache.get_namespace('import math')
    assert len(pystring_namespace_cache._cache) == 1

    contextparser_cache.get_context_parser('tests.arbpack.arbparser')
    assert len(contextparser_cache._cache) == 1

    step_cache.get_step('tests.arbpack.arbmutatingstep')
    assert len(step_cache._cache) == 1

    cache_admin.clear_all()

    assert backoff_cache._cache == builtin_backoffs
    assert file_cache._cache == {}
    assert loader_cache._cache == {}
    assert pystring_namespace_cache._cache == {}
    assert contextparser_cache._cache == {}
    assert step_cache._cache == {}



"""pypyr retry back-off strategies.

If you add a back-off strategy, be sure to add it also to the builtin_backoffs
attribute at the bottom.

Attributes:
    builtin_backoffs: dict mapping back-off name to implementation class.
"""
import abc
from collections import deque
import random


# region backoff base
class BackoffBase(abc.ABC):
    """Derive from me for a custom back-off strategy.

    Attributes:
        sleep (float): initial sleep interval in seconds.
        max_sleep (float): upper bound for sleep interval calculation.
        jrc (float): Jitter Range Coefficient.
        kwargs (dict): arbitrary keyword args for use in deriving classes.
    """

    def __init__(self, sleep=1, max_sleep=None, jrc=0, kwargs=None):
        """Initialize back-off strategy.

        Args:
            sleep (float): positive sleep interval in seconds.
            max_sleep (float): upper bound for sleep interval calculation.
            jrc (float): jitter randomizes between [sleep*jrc] and [sleep].
            kwargs (dict): arbitrary arguments for use by deriving classes.
        """
        self.sleep = sleep
        self.max_sleep = max_sleep
        self.jrc = jrc
        self.kwargs = kwargs

    @abc.abstractmethod
    def __call__(self, n):
        """Implement back-off sleep interval calculation here.

        Args:
            n (int): The iteration counter.

        Returns:
            Float. The sleep interval for iteration n.
        """
        raise NotImplementedError()  # pragma: no cover

    def min(self, sleep):
        """Return the smaller of sleep vs max_sleep."""
        max_sleep = self.max_sleep
        return min(sleep, max_sleep) if max_sleep else sleep

    def randomize(self, sleep):
        """Get random number in between (jrc*sleep) and (sleep).

        Does NOT guard against result < max_sleep. This is on purpose.
        It allows you to jitter under or over the sleep interval depending on
        if jrc > 1.
        """
        return random.uniform(sleep * self.jrc, sleep)

# endregion backoff base

# region backoff strategies implementation


class fixed(BackoffBase):
    """Fixed backoff interval(s).

    Sleep can be a single fixed interval, or a list with sleep intervals.

    In the case of a list, each retry iteration will use the next value in the
    list as the back-off interval, until it reaches the last item, which will
    repeat indefinitely.
    """

    def __init__(self, sleep, max_sleep=None, jrc=0, kwargs=None):
        """Initialize the fixed back-off strategy."""
        super().__init__(sleep, max_sleep, jrc, kwargs)
        if isinstance(sleep, (list, set)):
            self.queue = deque(sleep, len(sleep))
            self.fixed_sleep = self.min(self.queue[-1])
        else:
            self.queue = None
            self.fixed_sleep = self.min(sleep)

    def __call__(self, n):
        """Return the fixed backoff interval or the next item on list."""
        if self.queue:
            return self.min(self.queue.popleft())

        return self.fixed_sleep


class jitter(fixed):
    """Add jitter over fixed."""

    def __call__(self, n):
        """Randomize the fixed backoff interval."""
        return self.randomize(super().__call__(n))


class linear(BackoffBase):
    """Backoff increases linearly - the product of iteration and sleep."""

    def __call__(self, n):
        """Return the product of n * sleep."""
        return self.min(n * self.sleep)


class linearjitter(linear):
    """Add jitter over linear."""

    def __call__(self, n):
        """Randomize the product of n * sleep."""
        return self.randomize(super().__call__(n))


class exponential(BackoffBase):
    """Backoff with exponential growth of the sleep coefficient."""

    def __init__(self, sleep=1, max_sleep=None, jrc=0, kwargs=None):
        """Initialize the fixed back-off strategy."""
        super().__init__(sleep, max_sleep, jrc, kwargs)
        self.base = kwargs.get('base', 2) if kwargs else 2

    def __call__(self, n):
        """Return (base**n)*sleep."""
        return self.min(pow(self.base, n) * self.sleep)


class exponentialjitter(exponential):
    """Add jitter over exponential."""

    def __call__(self, n):
        """Randomize exponential result of n."""
        return self.randomize(super().__call__(n))

# endregion backoff strategies implementation


builtin_backoffs = {
    'fixed': fixed,
    'jitter': jitter,
    'linear': linear,
    'linearjitter': linearjitter,
    'exponential': exponential,
    'exponentialjitter': exponentialjitter
}



"""tar.py unit tests."""
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.tar
import pytest
from unittest.mock import patch, DEFAULT

# ------------------------- get file mode ------------------------------------#


def test_get_file_mode_for_reading():
    """File Mode for reading passes, with defaults."""
    # if there's no tarFormat, r:*
    assert 'r:*' == pypyr.steps.tar.get_file_mode_for_reading({})

    # if there's a tarFormat, r:blah
    assert 'r:blah' == pypyr.steps.tar.get_file_mode_for_reading(
        {'format': 'blah'})


def test_get_file_mode_for_writing():
    """File Mode for writing passes, with defaults."""
    # if there's no tarFormat, r:*
    assert 'w:xz' == pypyr.steps.tar.get_file_mode_for_writing({})

    # if there's a tarFormat, r:blah
    assert 'w:blah' == pypyr.steps.tar.get_file_mode_for_writing(
        {'format': 'blah'})

# ------------------------- get file mode ------------------------------------#


# ------------------------- tar base -----------------------------------------#


def test_tar_throws_on_empty_context():
    """Context must exist."""
    with pytest.raises(AssertionError) as err_info:
        pypyr.steps.tar.run_step(None)

    assert str(err_info.value) == "context must have value for pypyr.steps.tar"


def test_tar_throws_if_all_tar_context_missing():
    """Tar context keys must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.tar.run_step(Context({'arbkey': 'arbvalue'}))

    assert str(err_info.value) == ("context['tar'] doesn't exist. It must "
                                   "exist for pypyr.steps.tar.")


def test_tar_throws_if_tar_context_wrong_type():
    """Tar context keys must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.tar.run_step(
            Context({'tar': {'extractX': 'it shouldnt be a string'}}))

    assert str(err_info.value) == (
        "pypyr.steps.tar must have either extract or archive specified under "
        "the tar key. Or both of these. It has neither.")


def test_tar_throws_if_tar_context_no_value():
    """Tar context keys must exist."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.tar.run_step(
            Context({'tar': {'extract': None,
                             'archive': None}}))

    assert str(err_info.value) == (
        "pypyr.steps.tar must have either extract or archive specified under "
        "the tar key. Or both of these. It has neither.")


def test_tar_only_calls_extract():
    """Only calls extract if only extract specified."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'extract': [
            {'in': 'key2',
             'out': 'ARB_GET_ME1'},
            {'in': 'key4',
             'out': 'ARB_GET_ME2'}
        ]}
    })

    with patch.multiple('pypyr.steps.tar',
                        tar_archive=DEFAULT,
                        tar_extract=DEFAULT
                        ) as mock_tar:
        pypyr.steps.tar.run_step(context)

    mock_tar['tar_extract'].assert_called_once()
    mock_tar['tar_archive'].assert_not_called()


def test_tar_only_calls_archive():
    """Only calls archive if only archive specified."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'archive': [
            {'in': 'key2',
             'out': 'ARB_GET_ME1'},
            {'in': 'key4',
             'out': 'ARB_GET_ME2'}
        ]}
    })

    with patch.multiple('pypyr.steps.tar',
                        tar_archive=DEFAULT,
                        tar_extract=DEFAULT
                        ) as mock_tar:
        pypyr.steps.tar.run_step(context)

    mock_tar['tar_extract'].assert_not_called()
    mock_tar['tar_archive'].assert_called_once()


def test_tar_calls_archive_and_extract():
    """Calls both extract and archive when both specified."""
    context = Context({
        'key2': 'value2',
        'key1': 'value1',
        'key3': 'value3',
        'tar': {'archive': [
            {'in': 'key2',
             'out': 'ARB_GET_ME1'},
            {'in': 'key4',
             'out': 'ARB_GET_ME2'}
        ],
            'extract': [
            {'in': 'key2',
             'out': 'ARB_GET_ME1'},
            {'in': 'key4',
             'out': 'ARB_GET_ME2'}
        ]}
    })

    with patch.multiple('pypyr.steps.tar',
                        tar_archive=DEFAULT,
                        tar_extract=DEFAULT
                        ) as mock_tar:
        pypyr.steps.tar.run_step(context)

    mock_tar['tar_extract'].assert_called_once()
    mock_tar['tar_archive'].assert_called_once()


# ------------------------- tar base   ---------------------------------------#
#
# ------------------------- tar extract---------------------------------------#

def test_tar_extract_one_pass():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'extract': [
            {'in': './blah.tar.xz',
             'out': 'path/to/dir'}
        ]}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./blah.tar.xz', 'r:*')
    (mock_tarfile.return_value.
     __enter__().extractall.assert_called_once_with('path/to/dir'))


def test_tar_extract_one_pass_uncompressed():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'tar': {'extract': [
            {'in': './blah.tar.xz',
             'out': 'path/to/dir'}
        ],
            'format': ''}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./blah.tar.xz', 'r:')
    (mock_tarfile.return_value.
     __enter__().extractall.assert_called_once_with('path/to/dir'))


def test_tar_extract_one_with_interpolation():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'extract': [
            {'in': './{key3}.tar.xz',
             'out': 'path/{key2}/dir'}
        ]}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./value3.tar.xz', 'r:*')
    (mock_tarfile.return_value.
     __enter__().extractall.assert_called_once_with('path/value2/dir'))


def test_tar_extract_pass():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'extract': [
            {'in': './blah.tar.xz',
             'out': 'path/to/dir'},
            {'in': '/tra/la/la.tar.xz',
             'out': '.'}
        ]}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    assert mock_tarfile.call_count == 2
    mock_tarfile.assert_any_call('./blah.tar.xz', 'r:*')
    mock_tarfile.assert_any_call('/tra/la/la.tar.xz', 'r:*')
    (mock_tarfile.return_value.
     __enter__().extractall.call_count == 2)
    (mock_tarfile.return_value.
     __enter__().extractall.assert_any_call('path/to/dir'))
    (mock_tarfile.return_value.
     __enter__().extractall.assert_any_call('.'))

# ------------------------- tar extract---------------------------------------#

# ------------------------- tar archive---------------------------------------#


def test_tar_archive_one_pass():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'archive': [
            {'in': 'path/to/dir',
             'out': './blah.tar.xz'}
        ]}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./blah.tar.xz', 'w:xz')
    (mock_tarfile.return_value.
     __enter__().add.assert_called_once_with('path/to/dir', arcname='.'))


def test_tar_archive_one_pass_with_interpolation():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'tar': {'archive': [
            {'in': '{key2}/to/dir',
             'out': './blah.tar.{key1}'}
        ]}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./blah.tar.value1', 'w:xz')
    (mock_tarfile.return_value.
     __enter__().add.assert_called_once_with('value2/to/dir', arcname='.'))


def test_tar_archive_one_pass_without_compression():
    """Tar extract success case without compression."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'tar': {'archive': [
            {'in': 'path/to/dir',
             'out': './blah.tar.xz'}
        ],
            'format': ''}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./blah.tar.xz', 'w:')
    (mock_tarfile.return_value.
     __enter__().add.assert_called_once_with('path/to/dir', arcname='.'))


def test_tar_archive_one_pass_with_gz():
    """Tar extract success case with gz."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'tarFormat': 'gz',
        'tar': {'archive': [
            {'in': 'path/to/dir',
             'out': './blah.tar.gz'}
        ],
            'format': '{tarFormat}'}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    mock_tarfile.assert_called_once_with('./blah.tar.gz', 'w:gz')
    (mock_tarfile.return_value.
     __enter__().add.assert_called_once_with('path/to/dir', arcname='.'))


def test_tar_archive_pass():
    """Tar extract success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'blah',
        'tar': {'archive': [
            {'in': 'path/to/dir',
             'out': './{key3}.tar.xz'},
            {'in': '.',
             'out': '/tra/la/la.tar.xz'}
        ]}
    })

    with patch('tarfile.open') as mock_tarfile:
        pypyr.steps.tar.run_step(context)

    assert mock_tarfile.call_count == 2
    mock_tarfile.assert_any_call('./blah.tar.xz', 'w:xz')
    mock_tarfile.assert_any_call('/tra/la/la.tar.xz', 'w:xz')
    (mock_tarfile.return_value.
     __enter__().add.call_count == 2)
    (mock_tarfile.return_value.
     __enter__().add.assert_any_call('path/to/dir', arcname='.'))
    (mock_tarfile.return_value.
     __enter__().add.assert_any_call('.', arcname='.'))

# ------------------------- tar archive --------------------------------------#



"""Configuration settings.

Attributes:
    config: Use this global singleton from elsewere to access config values.
"""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Mapping
from io import StringIO
import os
from pathlib import Path
import sys
from typing import Any, Callable

import ruamel.yaml

import pypyr.errors
import pypyr.platform
import pypyr.toml
from pypyr.utils.types import cast_str_to_bool

CWD = Path.cwd()


class Config():
    """Configuration properties for pypyr.

    Look here for 'constant' values or configurable values.

    You probably shouldn't be instantiating this class directly yourself, use
    the global `config` attribute of this module instead, which contains a
    shared instance of this class - `from pypyr.config import config`.

    The standard constructor initializes the class with everything pypyr needs
    to run by default. Optionally, you can then also invoke init() on the
    instance to do heavier configuration setup that looks for the various
    config files on the filesystem, parses yaml+toml and reads the available
    values into the Config instance.

    NB: When adding attributes or properties to this class, remember to update
    Config.all_writable_props AND the custom __str__() overload also.

    Attributes:
        cwd: Path. The current working directory.
        json_indent: int. Indent for json output files.
        json_ascii: bool. If True escapes non-ascii chars.
        pipeline_subdir: str. 2nd pipeline look-up dir - subdir of cwd.
        log_config: dict. Logging config dict for logging.config.dictConfig.
        log_date_format: str. Format str for datetime in log output.
        log_notify_format: str. Format str for default log output.
        log_detail_format: str. Format str for detailed log output.
        default_backoff: str. Default retry back-off algorithm.
        default_cmd_encoding: str. Encoding to use on stdout/stderr with
            subprocess via pypyr.steps.cmd, pypyr.steps.cmds & shell.
        default_encoding: str. Encoding to use when working with files. None
            means to use the system defaults (utf-8, except for Windows.)
        default_loader: str. Default pipeline loader - 'pypyr.loaders.file'
        default_group: str. Step group name to run as entry-point - 'steps'.
        default_success_group: str. Name of step group to run on success -
                               'on_success'.
        default_failure_group: str. Name of step group to run on failure -
                               'on_failure'.
        shortcuts: dict. Pipeline run instructions with their inputs.
            Set by init().
        vars: dict. User provided variables to write into the pypyr context.
            Set by init().
        no_cache: bool. Default False. Bypass all pypyr caches entirely.
        platform_paths: pypyr.platform.PlatformPaths: O/S specific paths to
            config files & data dirs. Set by init().
        pyproject_toml: dict. The pyproject.toml file as a dict in a full.
            Set by init().
        skip_init: bool. Default False. Skip the init() method if this set.
        config_loaded_paths: list[Path]. List of paths of the loaded config
            files, in the order loaded.
    """

    all_writable_props = {
        'json_ascii',
        'json_indent',
        'pipelines_subdir',
        # logging
        'log_config',
        'log_date_format',
        'log_notify_format',
        'log_detail_format',
        # defaults
        'default_backoff',
        'default_cmd_encoding',
        'default_encoding',
        'default_loader',
        'default_group',
        'default_success_group',
        'default_failure_group',
        # flags
        'no_cache',
        # functional
        'shortcuts',
        'vars'}
    dict_props = {'shortcuts', 'vars'}
    scalar_props = all_writable_props - dict_props

    def __init__(self) -> None:
        """Initialize config with light-weight standard values."""
        # region writeable attributes

        # When adding an attr here, MUST add to Config.all_writable_props too.
        # Remember to amend custom __str__ overload when adding props/attrs.

        self.json_ascii = False
        self.json_indent = 2
        self.pipelines_subdir = 'pipelines'

        # logging
        self.log_config = None
        self.log_date_format = '%Y-%m-%d %H:%M:%S'
        self.log_notify_format = '%(message)s'
        self.log_detail_format = (
            '%(asctime)s %(levelname)s:%(name)s:%(funcName)s: %(message)s')

        # defaults
        self.default_backoff = 'fixed'
        # 'utf-8' is a magic string specially optimized in CPython
        self.default_cmd_encoding: str | None = os.getenv('PYPYR_CMD_ENCODING',
                                                          None)
        self.default_encoding: str | None = os.getenv('PYPYR_ENCODING', None)
        self.default_loader = 'pypyr.loaders.file'
        self.default_group = 'steps'
        self.default_success_group = 'on_success'
        self.default_failure_group = 'on_failure'

        # flags
        self.no_cache: bool = cast_str_to_bool(os.getenv('PYPYR_NO_CACHE',
                                                         '0'))

        # functional
        self.shortcuts: dict = {}
        self.vars: dict = {}

        # endregion writeable attributes

        # readonly properties set by default/light init
        current_platform = sys.platform
        self._platform: str = current_platform
        self._is_macos: bool = current_platform == 'darwin'
        self._is_windows: bool = current_platform == 'win32'
        self._is_posix: bool = not (self._is_macos or self._is_windows)

        # not writeable from config - readonly properties set by self.init()
        self._platform_paths: pypyr.platform.PlatformPaths | None = None
        self._pyproject_toml: dict | None = None
        self._skip_init = False
        self._config_loaded_paths: list[Path] = []

    @property
    def config_loaded_paths(self) -> list[Path] | None:
        """List of all the found config files loaded into runtime config.

        In the order that pypyr loaded the files.

        Set by init(). If init() didn't run, this is None.
        """
        return self._config_loaded_paths

    @property
    def cwd(self) -> Path:
        """Return the current working directory."""
        return CWD

    @property
    def is_macos(self) -> bool:
        """Is True if current platform is MacOS."""
        return self._is_macos

    @property
    def is_posix(self) -> bool:
        """Is True if current platform is POSIX."""
        return self._is_posix

    @property
    def is_windows(self) -> bool:
        """Is True if current platform is MacOS."""
        return self._is_windows

    @property
    def platform(self) -> str:
        """Platform (O/S) identifier."""
        return self._platform

    @property
    def platform_paths(self) -> pypyr.platform.PlatformPaths | None:
        """Platform specific paths & application directories.

        Set by init(). If init() didn't run, this is None. If no config files
        were found, this is None.
        """
        return self._platform_paths

    @property
    def pyproject_toml(self) -> dict | None:
        """Contents of ./pyproject_toml, if it exists.

        Set by init(). If init() didn't run, this is None. If init() ran but
        ./pyproject.toml didn't exist, also none.
        """
        return self._pyproject_toml

    @property
    def skip_init(self) -> bool:
        """Skip heavy init() - set by $PYPYR_SKIP_INIT."""
        return self._skip_init

    def init(self) -> None:
        """Heavy initialization to look-up config files on filesystem.

        Only call this if you want to load configuration settings from the
        config file look-up sequence. This is entirely optional - pypyr works
        with sensible defaults even when you don't want to call this.

        Order of precedence:
        1. /pypyr-config.yaml
        2. ./pyproject.toml
        3. Use $PYPYR_CONFIG_GLOBAL if available. If not, go to 4 & 5.
        4. $XDG_CONFIG_HOME - ~/.config/pypyr/config.yaml
        5. $XDG_CONFIG_DIRS/pypyr/config.yaml

        If $PYPYR_SKIP_INIT is 'true' or '1', this method returns without doing
        anything.
        """
        if cast_str_to_bool(os.getenv('PYPYR_SKIP_INIT', '0')):
            self._skip_init = True
            return

        # The final merge sticks, therefore go in reverse order 5 -> 1.
        env_config_path_str = os.getenv('PYPYR_CONFIG_GLOBAL', None)

        if env_config_path_str:
            # 3. Use $PYPYR_CONFIG_GLOBAL if available. If not, go to 4 & 5.
            env_config_path = Path(env_config_path_str)
            # Since user explicitly set this path, raise error if not exist.
            self.handle_path(env_config_path, raise_not_found=True)
            self._platform_paths = pypyr.platform.PlatformPaths(
                config_user=env_config_path,
                config_common=[env_config_path],
                data_dir_user=env_config_path,
                data_dir_common=[env_config_path])
        else:
            platform_paths = pypyr.platform.get_platform_paths('pypyr',
                                                               'config.yaml')

            self._platform_paths = platform_paths

            # 5. $XDG_CONFIG_DIRS/pypyr/config.yaml
            for path in reversed(platform_paths.config_common):
                # going in reverse order because the last update in the
                # sequence is the prevailing value.
                self.handle_path(path)

            # 4. $XDG_CONFIG_HOME - ~/.config/pypyr/config.yaml
            self.handle_path(platform_paths.config_user)

        # 2. ./pyproject.toml
        self.handle_path(Path('pyproject.toml'), self.load_pyproject_toml)

        # 1. /pypyr-config.yaml
        config_file_name = os.getenv('PYPYR_CONFIG_LOCAL',
                                     'pypyr-config.yaml')

        self.handle_path(Path(config_file_name))

    def handle_path(self,
                    path: Path,
                    handler: Callable | None = None,
                    raise_not_found: bool = False) -> None:
        """Load configuration file from path and merge into runtime config.

        handler signature is:
            handler(path: Path, raise_not_found: bool) -> Mapping

        If handler not specified, defaults to yaml loader.

        Args:
            path (Path): Path to file to load.
            handler (Callable): Callable that loads file and returns payload.
                                Defaults to yaml loader if not set.
            raise_not_found (bool): Raise error if file not found. Defaults
                                    to False.
        """
        payload = (handler(path, raise_not_found) if handler
                   else self.load_yaml(path, raise_not_found))

        if payload:
            if not isinstance(payload, Mapping):
                raise pypyr.errors.ConfigError(
                    f'Config file {path} should be a mapping (i.e a dict or '
                    'table) at the top level.')

            self.update(payload)
            self._config_loaded_paths.append(path)

    def update(self, input: Mapping) -> None:
        """Update self from input dict.

        Merges input dict into the current config instance (self).

        Args:
            input (Mapping): Merge this into the current instance.

        Returns: None
        """
        keys = input.keys()

        # 1. check all keys as expected
        difference = keys - Config.all_writable_props

        if difference:
            raise pypyr.errors.ConfigError(
                f'Unexpected config props: {difference}')

        # 2. update mapping keys
        dicts = keys & Config.dict_props
        for k in dicts:
            getattr(self, k).update(input[k])

        # 3. overwrite scalars
        scalars = keys & Config.scalar_props
        for k in scalars:
            setattr(self, k, input[k])

    def load_pyproject_toml(self, path: Path | str,
                            raise_error=False) -> Mapping | None:
        """Get pyproject toml config for pypyr.tool table.

        Args:
            path (Path): Path to file to load.
            raise_error (bool): Raise error if file not found. Defaults
                                to False.

        Returns:
            The [tool.pypyr] sub-table in the toml, if it exists.
        """
        try:
            toml = pypyr.toml.read_file(path)
        except OSError:
            if raise_error:
                raise pypyr.errors.ConfigError(
                    f'Could not open config file at {path}.') from OSError
            return None

        if toml:
            self._pyproject_toml = toml
            tool = toml.get('tool')
            if tool:
                return tool.get('pypyr')

        return None

    def load_yaml(self, path: Path | str, raise_error=False) -> Any:
        """Get config from yaml file.

        Args:
            path (Path): Path to file to load.
            raise_error (bool): Raise error if file not found. Defaults
                                to False.

        Returns:
            The yaml payload. Should be a Mapping, but could be any scalar or
            None.
        """
        parser = ruamel.yaml.YAML()
        try:
            # beg forgiveness later - mostly file WON'T be there.
            with open(path, encoding=self.default_encoding) as file:
                payload = parser.load(file)
        except OSError:
            if raise_error:
                raise pypyr.errors.ConfigError(
                    f'Could not open config file at {path}.') from OSError
            return None

        return payload

    def __str__(self) -> str:
        """Convert config to human friendly string."""
        # this might seem unnecessarily tedious - but neither pprint, nor dir()
        # or vars() give a consistent reliable listing of attributes + props
        # of interest.
        out = StringIO()
        yamler = ruamel.yaml.YAML()

        out.write('WRITEABLE PROPERTIES:\n\n')
        # instead of by hand creating str output, just dump yaml from dict
        # of all props
        writeable_props = {key: getattr(self, key)
                           for key in sorted(Config.all_writable_props)}

        yamler.dump(writeable_props, out)

        out.write('\n\nCOMPUTED PROPERTIES:\n\n')
        # the writeable props are all basic types, therefore the yaml dumper
        # knows what to do with them. The computed props use custom types,
        # which the yaml serializer can't handle without a custom serializer.
        # So either write a custom yaml serializer, or do it by hand. By hand
        # seemed less effort, comparatively.
        if self._config_loaded_paths:
            out.write('config_loaded_paths:\n')
            out.write(''.join(f'  - {k}\n' for k in self._config_loaded_paths))
        else:
            out.write('config_loaded_paths: []\n')

        out.write(f'cwd: {self.cwd}\n')

        out.write(f'is_macos: {self._is_macos}\n')
        out.write(f'is_posix: {self._is_posix}\n')
        out.write(f'is_windows: {self._is_windows}\n')
        out.write(f'platform: {self._platform}\n')

        if self._platform_paths:
            out.write(f'platform_paths:\n{self._platform_paths.to_str(2)}')
        else:
            out.write('platform_paths:\n')

        yamler.dump({'pyproject_toml': self._pyproject_toml}, out)

        out.write(f'skip_init: {self._skip_init}\n')
        return out.getvalue()


# global singleton of the config - use this to get your config values.
config = Config()



"""contextcopy.py unit tests."""
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.contextcopy
import pytest


def test_context_copy_throws_on_empty_context():
    """Input context must exist."""
    with pytest.raises(KeyNotInContextError):
        pypyr.steps.contextcopy.run_step(Context())


def test_context_copy_throws_on_contextcopy_missing():
    """Input contextcopy must exist in context."""
    with pytest.raises(KeyNotInContextError) as err_info:
        pypyr.steps.contextcopy.run_step(Context({'arbkey': 'arbvalue'}))

    assert str(err_info.value) == ("context['contextCopy'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.contextcopy.")


def test_context_copy_pass():
    """Input contextcopy success case."""
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'contextCopy': {
            'key2': 'key1',
            'key4': 'key3'
        }
    })

    pypyr.steps.contextcopy.run_step(context)

    assert context['key1'] == 'value1'
    assert context['key2'] == 'value1'
    assert context['key3'] == 'value3'
    assert context['key4'] == 'value3'



"""Initialize the pypyr package.

The most notable side-effect of importing this is package is that it adds a
NOTIFY (25) log level and notify() method to the global logger object.
"""
import pypyr.log.logger

__version__ = "5.9.1"

pypyr.log.logger.set_up_notify_log_level()



"""pypyrversion.py unit tests."""
import logging
import platform
from pypyr.context import Context
import pypyr.steps.pypyrversion
import pypyr.version
from tests.common.utils import patch_logger


def test_pypyr_version():
    """Return version as expected."""
    with patch_logger(
            'pypyr.steps.pypyrversion', logging.NOTIFY
    ) as mock_logger_notify:
        pypyr.steps.pypyrversion.run_step({})

    mock_logger_notify.assert_called_once_with(
        f'pypyr {pypyr.version.__version__} '
        f'python {platform.python_version()}')


def test_pypyr_version_context_out_same_as_in():
    """Context does not mutate."""
    context = Context({'test': 'value1'})
    pypyr.steps.pypyrversion.run_step(context)
    assert context['test'] == 'value1', "context not returned from step."



"""retry integration tests."""
from pathlib import Path
from unittest.mock import call, patch
from pypyr import pipelinerunner

working_dir_tests = Path(Path.cwd(), 'tests')


@patch('pypyr.retries.random.uniform', return_value=123)
@patch('time.sleep')
def test_retry_with_yaml_anchor(mock_sleep, mock_random):
    """Retry uses common shared anchors."""
    out = pipelinerunner.run(
        pipeline_name='tests/pipelines/retries/retry-anchors',
        dict_in={'outList': []})

    assert out['outList'] == ['s1.1',
                              's1.2',
                              's1.3',
                              's2.1',
                              's2.2',
                              's3.1',
                              's3.2',
                              's3.3']

    assert mock_random.mock_calls == [call(1.5, 3), call(4.25, 8.5)]
    assert mock_sleep.mock_calls == [call(2),
                                     call(4),
                                     call(2),
                                     call(123),
                                     call(123)]



"""pypyr step that writes payload out to a yaml file."""
import logging
from pathlib import Path

from pypyr.config import config
from pypyr.utils.asserts import assert_key_has_value
import pypyr.yaml

logger = logging.getLogger(__name__)

sentinel = object()


def run_step(context):
    """Write payload out to yaml file.

    If you do not set encoding, will use the system default, which is utf-8
    for everything except windows.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - fileWriteYaml
                    - path. mandatory. path-like. Write output file to
                      here. Will create directories in path for you.
                    - payload. optional. Write this to output file. If not
                      specified, output entire context.
                    - encoding. string. Defaults None (platform default,
                      usually 'utf-8').

    Returns:
        None.

    Raises:
        pypyr.errors.KeyNotInContextError: fileWriteYaml or
            fileWriteYaml['path'] missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileWriteYaml or
            fileWriteYaml['path'] exists but is None.

    """
    logger.debug("started")
    context.assert_key_has_value('fileWriteYaml', __name__)

    input_context = context.get_formatted('fileWriteYaml')
    assert_key_has_value(obj=input_context,
                         key='path',
                         caller=__name__,
                         parent='fileWriteYaml')
    out_path = Path(input_context['path'])
    # doing it like this to safeguard against accidentally dumping all context
    # with potentially sensitive values in it to disk if payload exists but is
    # None.
    payload = input_context.get('payload', sentinel)
    encoding = input_context.get('encoding', config.default_encoding)

    yaml_writer = pypyr.yaml.get_yaml_parser_roundtrip_for_context()

    logger.debug("opening destination file for writing: %s", out_path)

    out_path.parent.mkdir(parents=True, exist_ok=True)

    if payload is sentinel:
        payload = context.get_formatted_value(context)
    else:
        payload = input_context['payload']

    with open(out_path, 'w', encoding=encoding) as outfile:
        yaml_writer.dump(payload, outfile)

    logger.info("formatted context content and wrote to %s", out_path)
    logger.debug("done")



"""Create and manage stdlib venv aka virtual environments."""
# can remove __future__ once py 3.10 the lowest supported version
from __future__ import annotations
from collections.abc import Iterable, Mapping, Sequence
from concurrent.futures import Executor, Future
import logging
import os
from os import PathLike
import os.path
from pathlib import Path
import shlex
import subprocess
import sys
from types import SimpleNamespace
from venv import EnvBuilder

from pypyr.errors import ContextError, Error

logger = logging.getLogger(__name__)

CORE_VENV_DEPS = ('pip', 'setuptools')
SIMPLE_INPUT_TYPE = (str, PathLike)
IS_BEFORE_PY_3_9 = sys.version_info < (3, 9)


class VenvCreator():
    """Encapsulate python venv module functionality to create venvs."""

    DEFAULT_USE_SYMLINKS = False if os.name == 'nt' else True
    EXCEPTION_UPGRADE_CLEAR = """Error creating {path!s}.
You cannot set both upgrade and clear together for a venv."""

    EXCEPTION_NO_PIP = """Error creating {path!s}.
You cannot set `pip` or `upgrade_pip` when `with_pip` is False."""

    def __init__(self,
                 path: str | PathLike,
                 pip_extras: str = None,
                 system_site_packages: bool = False,
                 clear: bool = False,
                 symlinks: bool | None = None,
                 upgrade: bool = False,
                 with_pip: bool = True,
                 prompt: str = None,
                 upgrade_pip: bool = True,
                 is_quiet: bool = False) -> None:
        """Initialize VenvCreator."""
        if upgrade and clear:
            raise ContextError(
                VenvCreator.EXCEPTION_UPGRADE_CLEAR.format(path=path))

        if with_pip is False:
            upgrade_pip = False

            if pip_extras:
                raise ContextError(
                    VenvCreator.EXCEPTION_NO_PIP.format(path=path))

        self.path = str(Path(path).expanduser().resolve())
        self.pip_extras = pip_extras
        self.upgrade_pip = upgrade_pip
        self.with_pip = with_pip
        self.is_quiet = is_quiet
        if symlinks is None:
            symlinks = VenvCreator.DEFAULT_USE_SYMLINKS
            logger.debug("venv using symlinks: %s", symlinks)

        # env_builder always sets upgrade_deps False to prevent the pip install
        # running in parallel in ThreadPoolExecutor. If upgrade_pip is True,
        # will do the pip upgrade as part of install_dependencies in serial
        # AFTER the parallel env creation is complete.
        self._env_builder = EnvBuilderWithExtraDeps(
            system_site_packages=system_site_packages,
            clear=clear,
            symlinks=symlinks,
            upgrade=upgrade,
            with_pip=with_pip,
            prompt=prompt,
            upgrade_deps=False,
            is_quiet=is_quiet)

        self._future: Future | None = None
        self._is_done = False

    def check_result(self) -> Exception | None:
        """Return exception if any. None means success."""
        if self._future:
            ex = self._future.exception()
            if ex is None or isinstance(ex, Exception):
                return ex
            else:
                # raise BaseException immediately - KeyboardInterrupt etc.
                # not sure to what degree this is just mypy bleating over
                # nothing (future.exception typed as BaseException not
                # Exception.)
                raise ex
        else:
            raise Error(
                "You can only call check_result after create_in_executor. "
                + f"Error creating venv {self.path!s}"
            )

    def create(self) -> None:
        """Create the venv on disk.

        This method MUST be safe for parallel execution.

        So no pip install, unless pip in future support parallel execution.
        """
        logger.debug("started")
        self._env_builder.create(self.path)
        self._is_done = True
        logger.debug("done create venv path %s", self.path)

    def create_in_executor(self, executor: Executor) -> None:
        """Run create in the provided executor."""
        logger.debug("starting venv creation: %s", self.path)
        self._future = executor.submit(self.create)

    def install_dependencies(self) -> None:
        """Upgrade pip & install extra dependencies. Run me in serial."""
        logger.debug("started")
        if not self._is_done:
            # you really shouldn't end up here, unless API user is calling this
            # out of sequence, thus raise hard exception.
            raise Error(
                "you can't call install_dependencies before the environment "
                + f"is installed successfully for {self.path!s}.")

        env_builder = self._env_builder
        if not env_builder.with_pip:
            raise Error(
                f"you can't install extra dependencies into {self.path!s} "
                + "because with_pip is False.")

        if self.upgrade_pip:
            # upgrades pip & setuptools
            logger.debug("upgrading pip & setuptools in %s", self.path)
            env_builder.upgrade_dependencies(self._env_builder.context)

        if not self.pip_extras:
            logger.debug(
                "done - no extra dependencies to install for path %s",
                self.path)
            return

        env_builder.pip_install_extras(self.pip_extras)

        logger.debug("done")

    @classmethod
    def from_mapping(cls, mapping: Mapping) -> Iterable[VenvCreator]:
        """Create VenvCreator from a dict/mapping input.

        Args:
            mapping (Mapping): dict/mapping containing inputs for VenvCreator

        Returns:
            iterator where each item is a VenvCreator. Creates multiple
            instances when input `path` contains >1 path.
        """
        raw_dirs = mapping.get('path')

        if not raw_dirs:
            raise ContextError(
                "path is mandatory on mapping input for venv create.")

        if not isinstance(raw_dirs, (Sequence, set)):
            raise ContextError(
                "venv.path input should be a string or list. "
                + f"Current input is: {raw_dirs}")

        dirs = ([raw_dirs] if isinstance(
            raw_dirs, SIMPLE_INPUT_TYPE) else raw_dirs)

        system_site_packages = mapping.get('system_site_packages', False)
        clear = mapping.get('clear', False)
        symlinks = mapping.get('symlinks')
        upgrade = mapping.get('upgrade', False)
        with_pip = mapping.get('with_pip', True)
        prompt = mapping.get('prompt')
        upgrade_pip = mapping.get('upgrade_pip', True)
        pip_extras = mapping.get('pip')
        is_quiet = mapping.get('quiet', False)

        for path in dirs:
            yield cls(path=path,
                      pip_extras=pip_extras,
                      system_site_packages=system_site_packages,
                      clear=clear,
                      symlinks=symlinks,
                      upgrade=upgrade,
                      with_pip=with_pip,
                      prompt=prompt,
                      upgrade_pip=upgrade_pip,
                      is_quiet=is_quiet)


class EnvBuilderWithExtraDeps(EnvBuilder):
    """Derived stdlib EnvBuilder to persist in-flight context."""

    def __init__(self, **kwargs):
        """Initialize with special is_quiet flag."""
        self.is_quiet: bool = kwargs.pop('is_quiet', False)
        # the base ctor still does the actual work
        if IS_BEFORE_PY_3_9:
            # upgrade_deps only arrived in python 3.9
            self.upgrade_deps = kwargs.pop('upgrade_deps', False)

        super().__init__(**kwargs)

    def pip_install_extras(self, pip_args: str) -> None:
        """Run python -m pip install {pip_args} in venv.

        You can ONLY call this AFTER post_setup() has run, which in practice
        means you have to have called create().
        """
        logger.debug("installing extra dependencies for path %s",
                     self.context.env_dir)

        cmd = [self.context.env_exec_cmd, '-m', 'pip', 'install']
        if self.is_quiet:
            cmd.append('-q')

        # if input is a list already, no need to shlex split
        shlexed_deps = shlex.split(pip_args) if isinstance(pip_args,
                                                           str) else pip_args

        # each arg could contain a path, thus run expand on all args
        for i, v in enumerate(shlexed_deps):
            shlexed_deps[i] = os.path.expanduser(v)

        cmd.extend(shlexed_deps)

        logger.debug("running %s", cmd)
        subprocess.run(cmd, check=True)

        logger.debug("done")

    def post_setup(self, context: SimpleNamespace) -> None:
        """Save context - this is only way to get at it."""
        logger.debug("started - saving EnvBuilder context...")
        # the alternative to doing this is to override create(), which is
        # gnarly with a lot of logic, so rather not mess with it for
        # maintainability's sake.
        self.context = context
        logger.debug("done")

    def upgrade_dependencies(self, context: SimpleNamespace) -> None:
        """Upgrade pip & setuptools, with quiet switch added if specified."""
        logger.debug('Upgrading %s packages in %s',
                     CORE_VENV_DEPS,
                     context.bin_path)
        cmd = [context.env_exec_cmd, '-m', 'pip', 'install', '--upgrade']

        # this override is identical to base except for this quiet switch
        if self.is_quiet:
            cmd.append('-q')

        cmd.extend(CORE_VENV_DEPS)
        subprocess.check_call(cmd)



"""stop.py unit tests."""
import pytest
from pypyr.context import Context
from pypyr.errors import Stop
import pypyr.steps.stop


def test_step_stop():
    """Raise stop from stop."""
    with pytest.raises(Stop):
        pypyr.steps.stop.run_step({})


def test_step_stop_context_same():
    """Context endures on Stop."""
    context = Context({'test': 'value1'})
    with pytest.raises(Stop):
        pypyr.steps.stop.run_step(context)
    assert context['test'] == 'value1', "context not returned from step."



"""Unit tests for pypyr/platform.py."""
import os
from pathlib import Path
from unittest.mock import Mock, patch

import pytest

from pypyr import platform

HOME = Path.home()

# region PlatformPaths
cu = Path('/cu')
cc = Path('/cc')
du = Path('/du')
dc = Path('/dc')


def test_to_str_no_offset():
    """Str with zero offset."""
    pp = platform.PlatformPaths(config_user=cu,
                                config_common=[cc],
                                data_dir_user=du,
                                data_dir_common=[dc])
    assert str(pp) == (f'config_user: {cu}\n'
                       'config_common:\n'
                       f'  - {cc}\n'
                       f'data_dir_user: {du}\n'
                       'data_dir_common:\n'
                       f'  - {dc}\n'
                       )


def test_to_str_offset_2():
    """Str with offset 2."""
    pp = platform.PlatformPaths(config_user=cu,
                                config_common=[cc],
                                data_dir_user=du,
                                data_dir_common=[dc])
    assert pp.to_str(2) == (f'  config_user: {cu}\n'
                            '  config_common:\n'
                            f'    - {cc}\n'
                            f'  data_dir_user: {du}\n'
                            '  data_dir_common:\n'
                            f'    - {dc}\n'
                            )


def test_to_str_offset_3():
    """Str with offset 3."""
    cc2 = Path('cc2')
    pp = platform.PlatformPaths(config_user=cu,
                                config_common=[cc, cc2],
                                data_dir_user=du,
                                data_dir_common=[dc])
    assert pp.to_str(3) == (f'   config_user: {cu}\n'
                            '   config_common:\n'
                            f'     - {cc}\n'
                            f'     - {cc2}\n'
                            f'   data_dir_user: {du}\n'
                            '   data_dir_common:\n'
                            f'     - {dc}\n'
                            )

# endregion PlatformPaths

# region fixtures


@pytest.fixture
def no_xdg_envs(monkeypatch):
    """Unset XDG env vars."""
    monkeypatch.delenv('XDG_CONFIG_HOME', raising=False)
    monkeypatch.delenv('XDG_CONFIG_DIRS', raising=False)
    monkeypatch.delenv('XDG_DATA_HOME', raising=False)
    monkeypatch.delenv('XDG_DATA_DIRS', raising=False)


@pytest.fixture
def xdg_envs(monkeypatch):
    """Set XDG env vars."""
    monkeypatch.setenv('XDG_CONFIG_HOME', '/ch/')
    monkeypatch.setenv('XDG_CONFIG_DIRS', f'/cc{os.pathsep}/cc2')
    monkeypatch.setenv('XDG_DATA_HOME', '/dh')
    monkeypatch.setenv('XDG_DATA_DIRS', f'/dc/{os.pathsep}/dc2')


@pytest.fixture
def xdg_env_single(monkeypatch):
    """Set XDG env vars with single paths in the DIRS arrays."""
    monkeypatch.setenv('XDG_CONFIG_HOME', '/ch/')
    monkeypatch.setenv('XDG_CONFIG_DIRS', '/cc')
    monkeypatch.setenv('XDG_DATA_HOME', '/dh')
    monkeypatch.setenv('XDG_DATA_DIRS', '/dc/')


@pytest.fixture
def not_android(monkeypatch):
    """Unset Android $ENVs."""
    monkeypatch.delenv('ANDROID_DATA', raising=False)
    monkeypatch.delenv('ANDROID_ROOT', raising=False)


@pytest.fixture()
def env_android(monkeypatch):
    """Set $ENVs and path separator to mimic Android environment."""
    monkeypatch.setenv('ANDROID_DATA', '/data')
    monkeypatch.setenv('ANDROID_ROOT', '/system')
    monkeypatch.setattr('pypyr.platform.os.pathsep', ':')


@pytest.fixture()
def env_android_halfway(monkeypatch):
    """Set $ENVs to mimic half of the $ENVs for Android."""
    monkeypatch.setenv('ANDROID_DATA', '/data')
    monkeypatch.setenv('ANDROID_ROOT', '/arbxyzsdkskd')


@pytest.fixture
def windows(monkeypatch, not_android):
    """Set windows $ENV and pathsep."""
    monkeypatch.setenv('ALLUSERSPROFILE', 'C:/ProgramData1')
    monkeypatch.setattr('pypyr.platform.os.pathsep', ';')
    monkeypatch.setattr('pypyr.platform.sys.platform', 'win32')


@pytest.fixture
def posix(monkeypatch, no_win, not_android):
    """Set path separator to colon."""
    monkeypatch.setattr('pypyr.platform.os.pathsep', ':')


@pytest.fixture
def linux(monkeypatch, posix):
    """Set platform str to linux."""
    monkeypatch.setattr('pypyr.platform.sys.platform', 'linux')


@pytest.fixture
def cygwin(monkeypatch, posix):
    """Set platform str to cygwin."""
    monkeypatch.setattr('pypyr.platform.sys.platform', 'cygwin')


@pytest.fixture
def freebsd(monkeypatch, posix):
    """Set platform str to cygwin."""
    monkeypatch.setattr('pypyr.platform.sys.platform', 'freebsd')


@pytest.fixture
def darwin(monkeypatch, posix):
    """Set platform str to darwin."""
    monkeypatch.setattr('pypyr.platform.sys.platform', 'darwin')


@pytest.fixture
def no_win(monkeypatch):
    """Unset Windows ALLUSERSPROFILE environment variable."""
    monkeypatch.delenv('ALLUSERSPROFILE', raising=False)

# endregion

# region Xdg


def test_xdg_defaults(linux, no_xdg_envs):
    """Xdg with no $ENV set."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path(HOME, '.config/pypyr/config.yaml'),
        config_common=[Path('/etc/xdg/pypyr/config.yaml')],
        data_dir_user=Path(HOME, '.local/share/pypyr'),
        data_dir_common=[Path('/usr/local/share/pypyr'),
                         Path('/usr/share/pypyr')])


def test_xdg_vars_set(cygwin, xdg_envs):
    """Xdg with all $ENV set."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path('/ch//pypyr/config.yaml'),
        config_common=[Path('/cc/pypyr/config.yaml'),
                       Path('/cc2/pypyr/config.yaml')],
        data_dir_user=Path('/dh/pypyr'),
        data_dir_common=[Path('/dc/pypyr'),
                         Path('/dc2/pypyr')])


def test_xdg_vars_set_single(freebsd, xdg_env_single):
    """Xdg with all $ENV set and list values only have single values."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path('/ch//pypyr/config.yaml'),
        config_common=[Path('/cc/pypyr/config.yaml')],
        data_dir_user=Path('/dh/pypyr'),
        data_dir_common=[Path('/dc/pypyr')])

# endregion Xdg

# region MacOs


def test_mac_defaults(darwin, no_xdg_envs):
    """Mac with no xdg $ENV set."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path(HOME, '.config/pypyr/config.yaml'),
        config_common=[Path('/Library/Application Support/pypyr/config.yaml')],
        data_dir_user=Path(HOME, '.local/share/pypyr'),
        data_dir_common=[Path('/Library/Application Support/pypyr')])


def test_mac_vars_set(darwin, xdg_envs):
    """Xdg on Mac with all $ENV set."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path('/ch//pypyr/config.yaml'),
        config_common=[Path('/cc/pypyr/config.yaml'),
                       Path('/cc2/pypyr/config.yaml')],
        data_dir_user=Path('/dh/pypyr'),
        data_dir_common=[Path('/dc/pypyr'),
                         Path('/dc2/pypyr')])

# endregion MacOs

# region Windows


def test_win_defaults(windows, no_xdg_envs):
    """Windows with no xdg $ENV set."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path(HOME, '.config/pypyr/config.yaml'),
        config_common=[Path('C:/ProgramData1/pypyr/config.yaml')],
        data_dir_user=Path(HOME, '.local/share/pypyr'),
        data_dir_common=[Path('C:/ProgramData1/pypyr/')])


def test_win_allusersprofile_not_set(windows, no_win, no_xdg_envs):
    """Raise error when $ALLUSERSPROFILE is not set on Windows."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')
    assert pp == platform.PlatformPaths(
        config_user=Path(HOME, '.config/pypyr/config.yaml'),
        config_common=[Path('C:/ProgramData/pypyr/config.yaml')],
        data_dir_user=Path(HOME, '.local/share/pypyr'),
        data_dir_common=[Path('C:/ProgramData/pypyr/')])


def test_win_vars_set(windows, xdg_envs):
    """Win with all xdg $ENV set."""
    pp = platform.get_platform_paths('pypyr', 'config.yaml')

    assert pp == platform.PlatformPaths(
        config_user=Path('/ch//pypyr/config.yaml'),
        config_common=[Path('/cc/pypyr/config.yaml'),
                       Path('/cc2/pypyr/config.yaml')],
        data_dir_user=Path('/dh/pypyr'),
        data_dir_common=[Path('/dc/pypyr'),
                         Path('/dc2/pypyr')])
# endregion Windows

# region android


def get_jnius_mock():
    """Create a jnius mock that return path /arb."""
    mock_jnius = Mock()
    mock_autoclass = Mock()
    mock_get_abs_path = Mock(return_value='/arb')
    mock_get_parent = Mock()
    mock_get_parent.return_value.getAbsolutePath = mock_get_abs_path
    mock_get_files_dir = Mock()
    mock_get_files_dir.return_value.getParentFile = mock_get_parent

    mock_autoclass.return_value.getFilesDir = mock_get_files_dir

    mock_jnius.autoclass = mock_autoclass
    return mock_jnius


def test_android_defaults(env_android):
    """Android with jnius success path."""
    mock_jnius = get_jnius_mock()

    with patch.dict('sys.modules', {'jnius': mock_jnius}):
        pp = platform.get_platform_paths('pypyr', 'config.yaml')

    mock_jnius.autoclass.assert_called_once_with('android.content.Context')

    assert pp == platform.PlatformPaths(
        config_user=Path('/arb/shared_prefs/pypyr/config.yaml'),
        config_common=[Path('/arb/shared_prefs/pypyr/config.yaml')],
        data_dir_user=Path('/arb/files/pypyr'),
        data_dir_common=[Path('/arb/files/pypyr')])


@patch('pypyr.platform.sys')
def test_android_not_found(mock_sys, env_android):
    """Android with jnius failed import."""
    mock_jnius = get_jnius_mock()
    mock_jnius.autoclass.side_effect = ModuleNotFoundError()

    mock_sys.path = ['/arb', '/hello']
    with patch.dict('sys.modules', {'jnius': mock_jnius}):
        with pytest.raises(OSError) as err:
            platform.get_platform_paths('pypyr', 'config.yaml')

    assert str(err.value) == "Cannot find path to android app folder"
    mock_jnius.autoclass.assert_called_once_with('android.content.Context')


@patch('pypyr.platform.sys')
def test_android_via_sys_path(mock_sys, env_android):
    """Android with jnius failed import finds dir via sys.path."""
    mock_jnius = get_jnius_mock()
    mock_jnius.autoclass.side_effect = ModuleNotFoundError()

    mock_sys.path = ['/arb', '/data/user/1/a/files']
    with patch.dict('sys.modules', {'jnius': mock_jnius}):
        pp = platform.get_platform_paths('pypyr', 'config.yaml')

    mock_jnius.autoclass.assert_called_once_with('android.content.Context')

    assert pp == platform.PlatformPaths(
        config_user=Path('/data/user/1/a/shared_prefs/pypyr/config.yaml'),
        config_common=[Path('/data/user/1/a/shared_prefs/pypyr/config.yaml')],
        data_dir_user=Path('/data/user/1/a/files/pypyr'),
        data_dir_common=[Path('/data/user/1/a/files/pypyr')])


@patch('pypyr.platform.sys')
def test_android_halfway(mock_sys, env_android_halfway):
    """Android has to have both envs matching to pass."""
    mock_sys.platform = 'win32'
    assert platform.get_platform_dir_finder() is platform.Windows

# endregion android



"""fileformat.py integration tests."""
import logging
import os
import tempfile
from pathlib import Path
from unittest.mock import call, patch

import pypyr.utils.filesystem as filesystem
import pytest
from pypyr.errors import Error
# region setup/teardown/fixtures
from tests.common.utils import patch_logger


@pytest.fixture(scope="module")
def temp_dir():
    """Make tmp dir in testfiles/out."""
    # runs at start of module
    os.makedirs('./tests/testfiles/out/', exist_ok=True)
    tmp_dir = tempfile.TemporaryDirectory(dir='./tests/testfiles/out/')
    yield Path(tmp_dir.name)
    # runs when module done
    tmp_dir.cleanup()
    os.rmdir('./tests/testfiles/out/')


@pytest.fixture
def temp_file_creator(temp_dir):
    """Create a function that makes a temp file in a fixture."""
    temp_files = []

    def create_temp_file():
        """Create a temp file for each function invocation."""
        # close file handle to allow other code to use file on windows
        with tempfile.NamedTemporaryFile(dir=temp_dir, delete=False) as tmp:
            p = Path(tmp.name)

        temp_files.append(p)
        return p

    yield create_temp_file

    # this runs after the decorated test goes out of scope
    for file in temp_files:
        try:
            # can't use missing_ok=True until py 3.8 the min ver.
            file.unlink()
        except FileNotFoundError:
            # just clean-up, if file already been removed no probs
            pass

# endregion setup/teardown/fixtures

# region FileRewriter


class ArbRewriter(filesystem.FileRewriter):
    """Derived FileRewriter useful for capturing test inputs."""

    def __init__(self, formatter):
        """Initialize calls list."""
        self.in_out_calls = []
        self.in_out_calls_paths = []
        super().__init__(formatter)

    def in_to_out(self, in_path, out_path=None):
        """Don't do much except log inputs."""
        self.in_out_calls.append({'in': in_path, 'out': out_path})
        out_obj = Path(out_path).resolve() if out_path else None
        self.in_out_calls_paths.append({'in': Path(in_path).resolve(),
                                        'out': out_obj})

    def assert_in_to_out_call(self, in_path, out_path):
        """Assert that the in/out pair exists in the call list."""
        assert (any(dict for dict in self.in_out_calls
                    if dict['in'] == in_path
                    and dict['out'] == out_path))

    def assert_in_to_out_call_path(self, in_path, out_path):
        """Assert that the in/out paths exists in the call list.

        Args:
            in_path: Pathlib Path. Uses .resolve() to normalize path.
            out_path: Pathlib Path. Uses .resolve() to normalize path.

        Returns:
            None.

        Raises:
            AssertionError if no match found.

        """
        in_resolved = Path(in_path).resolve() if in_path else None
        out_resolved = Path(out_path).resolve() if out_path else None
        assert (any(dict for dict in self.in_out_calls_paths
                    if dict['in'] == in_resolved
                    and dict['out'] == out_resolved))

    def assert_in_to_out_call_count(self, count):
        """Assert that the number of in_to_out calls is count."""
        assert len(self.in_out_calls) == count


def test_arbwriter_assert_in_to_out_call_exists():
    """Arb Rewriter asserts should work as advertized."""
    arb = ArbRewriter('formatter')
    arb.in_to_out('1', '2')
    arb.in_to_out('3', '4')

    arb.assert_in_to_out_call('1', '2')
    arb.assert_in_to_out_call('3', '4')
    arb.assert_in_to_out_call_path(Path('1'), Path('2'))
    arb.assert_in_to_out_call_path(Path('3'), Path('4'))
    arb.assert_in_to_out_call_count(2)

    with pytest.raises(AssertionError):
        arb.assert_in_to_out_call('5', '6')


def test_filerewriter_files_in_to_out_no_in_found_no_out():
    """Files in to out does nothing when in glob finds nothing, no outpath."""
    rewriter = ArbRewriter("formatter")

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out('./arb/*')

    assert mock_logger_info.mock_calls == [
        call("./arb/* found no files")]

    assert not rewriter.in_out_calls


def test_filerewriter_files_in_to_out_no_in_found_with_out():
    """Files in to out does nothing when in glob finds nothing, with out."""
    rewriter = ArbRewriter("formatter")

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out('./arb/*', './arb2/*')

    assert mock_logger_info.mock_calls == [
        call("./arb/* found no files")]

    assert not rewriter.in_out_calls


def test_filerewriter_files_in_to_out_1_to_1_str(temp_dir,
                                                 temp_file_creator):
    """Files in to out invokes file_in_to_out for each file on str input."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()
    out = temp_dir.joinpath('outfile')

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(str(file1), str(out))

    assert mock_logger_info.mock_calls == [
        call(f"read {file1}, formatted and wrote 1 file(s) to {out}")]

    assert rewriter.in_out_calls == [{'in': file1, 'out': out}]


def test_filerewriter_files_in_to_out_1_to_1_path(temp_dir,
                                                  temp_file_creator):
    """Files in to out invokes file_in_to_out for each file on path input."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()
    out = temp_dir.joinpath('outfile')

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(file1, out)

    assert mock_logger_info.mock_calls == [
        call(f"read {file1}, formatted and wrote 1 file(s) to {out}")]

    assert rewriter.in_out_calls == [{'in': file1, 'out': out}]


def test_filerewriter_files_in_to_out_3_to_1(temp_dir,
                                             temp_file_creator):
    """Files in to out can't write multiple ins to 1 out."""
    rewriter = ArbRewriter("formatter")

    temp_file_creator()
    temp_file_creator()
    temp_file_creator()

    in_glob = temp_dir.joinpath('*')
    out = temp_dir.joinpath('outfile')

    with pytest.raises(Error) as err:
        rewriter.files_in_to_out(in_glob, out)

    assert str(err.value) == (
        f'{in_glob} resolves to 3 files, but you '
        f'specified only a single file as out {out}. If the outpath is meant '
        'to be a directory, put a / at the end.')

    assert not rewriter.in_out_calls


def test_filerewriter_files_in_to_out_edit_dir_slash(temp_dir,
                                                     temp_file_creator):
    """Files in to out edits when out existing dir shown by /."""
    rewriter = ArbRewriter("formatter")

    temp_file_creator()
    temp_file_creator()

    in_path = temp_dir.joinpath('*')
    out = str(temp_dir) + os.sep

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(in_path, out)

    assert mock_logger_info.mock_calls == [
        call(f"read {in_path}, formatted and wrote 2 file(s) to {out}")]

    rewriter.assert_in_to_out_call_count(2)


def test_filerewriter_files_in_to_out_dir_slash_create(temp_dir,
                                                       temp_file_creator):
    """Files in to out when out dir shown by / creates dir."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()
    file2 = temp_file_creator()

    in_path = temp_dir.joinpath("*")
    out = str(temp_dir.joinpath('sub')) + os.sep

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(in_path, out)

    assert mock_logger_info.mock_calls == [
        call(f"read {in_path}, formatted and wrote 2 file(s) to {out}")]

    path_out = Path(out)
    # out/sub exists
    assert path_out.is_dir()
    rewriter.assert_in_to_out_call_count(2)
    # outfile created with same name as in-file in out/sub dir
    rewriter.assert_in_to_out_call_path(file1, path_out.joinpath(file1.name))
    rewriter.assert_in_to_out_call_path(file2, path_out.joinpath(file2.name))


def test_filerewriter_files_in_to_out_edit_dir_no_slash(temp_dir,
                                                        temp_file_creator):
    """Files in to out edits when out existing dir, without trailing slash."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(file1, temp_dir)

    assert mock_logger_info.mock_calls == [
        call(f"read {file1}, formatted and wrote 1 file(s) to {temp_dir}")]

    rewriter.assert_in_to_out_call_count(1)
    rewriter.assert_in_to_out_call_path(Path(file1),
                                        temp_dir.joinpath(file1.name))


def test_filerewriter_files_in_to_out_no_out(temp_file_creator):
    """Files in to out with no out edits in place."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(file1)

    assert mock_logger_info.mock_calls == [
        call(f"edited & wrote 1 file(s) at {file1}")]

    rewriter.assert_in_to_out_call_count(1)
    rewriter.assert_in_to_out_call_path(Path(file1), None)


def test_filerewriter_files_in_to_out_in_glob_no_out(temp_dir,
                                                     temp_file_creator):
    """Files in to out with in glob and no out edits in place."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()
    file2 = temp_file_creator()
    file3 = temp_file_creator()

    in_glob = temp_dir.joinpath('*')

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(in_glob)

    assert mock_logger_info.mock_calls == [
        call(f"edited & wrote 3 file(s) at {in_glob}")]

    rewriter.assert_in_to_out_call_count(3)
    rewriter.assert_in_to_out_call_path(file1, None)
    rewriter.assert_in_to_out_call_path(file2, None)
    rewriter.assert_in_to_out_call_path(file3, None)


def test_filerewriter_files_in_to_out_in_list_no_out(temp_file_creator):
    """Files in to out with in list and no out edits in place."""
    rewriter = ArbRewriter("formatter")

    file1 = temp_file_creator()
    file2 = temp_file_creator()
    file3 = temp_file_creator()

    in_list = [file1, str(file2), file3]

    with patch_logger(
            'pypyr.utils.filesystem', logging.INFO
    ) as mock_logger_info:
        rewriter.files_in_to_out(in_list)

    assert mock_logger_info.mock_calls == [
        call(f"edited & wrote 3 file(s) at {in_list}")]

    rewriter.assert_in_to_out_call_count(3)
    rewriter.assert_in_to_out_call_path(file1, None)
    rewriter.assert_in_to_out_call_path(file2, None)
    rewriter.assert_in_to_out_call_path(file3, None)

# endregion FileRewriter

# region ObjectRewriter


class ArbRepresenter(filesystem.ObjectRepresenter):
    """Test representer."""

    def __init__(self, is_binary=False):
        """Initialize me."""
        super().__init__(is_binary)
        self.load_payload = None
        self.dump_payload = None
        self.load_call_count = 0
        self.dump_call_count = 0

    def load(self, file):
        """Get payload from file."""
        self.load_payload = file.read()
        self.load_call_count += 1
        return self.load_payload

    def dump(self, file, payload):
        """Write payload to file."""
        file.write(payload)
        self.dump_payload = payload
        self.dump_call_count += 1


def get_arb_formatted(input):
    """Arbitrary formatter takes a string and returns a string."""
    return f'X{input}X'


def get_binary_formatted(input):
    """Arbitrary formatter takes bytes and concats some bytes."""
    return b'X' + input + b'X'


def get_arb_formatted_iter(input):
    """Arbitrary formatter takes an iterable and returns an iterable."""
    for item in input:
        yield f'X{item}X'


def test_objectrewriter_in_to_out_str(temp_dir, temp_file_creator):
    """Object Rewriter writes single in to out str."""
    representer = ArbRepresenter()
    rewriter = filesystem.ObjectRewriter(get_arb_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')
    out_str = str(temp_dir.joinpath('sub', 'filename'))

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(str(path_in), out_str)

    assert representer.load_payload == 'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == 'XyyyX'
    assert representer.dump_call_count == 1

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call(f"opening destination file for writing: {out_str}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'yyy'
    path_out = Path(out_str)
    assert path_out.is_file()
    assert path_out.read_text() == 'XyyyX'


def test_objectrewriter_in_to_out_path(temp_dir, temp_file_creator):
    """Object Rewriter writes single in to out path object."""
    representer = ArbRepresenter()
    rewriter = filesystem.ObjectRewriter(get_arb_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')
    path_out = temp_dir.joinpath('sub', 'filename')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(path_in, path_out)

    assert representer.load_payload == 'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == 'XyyyX'
    assert representer.dump_call_count == 1

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call(f"opening destination file for writing: {path_out}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'yyy'
    assert path_out.is_file()
    assert path_out.read_text() == 'XyyyX'


def test_objectrewriter_in_to_out_path_binary(temp_dir, temp_file_creator):
    """Object Rewriter writes single in to out path object in binary mode."""
    representer = ArbRepresenter(True)
    rewriter = filesystem.ObjectRewriter(get_binary_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')
    path_out = temp_dir.joinpath('sub', 'filename')

    rewriter.in_to_out(path_in, path_out)

    assert representer.load_payload == b'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == b'XyyyX'
    assert representer.dump_call_count == 1

    assert path_in.is_file()
    assert path_in.read_text() == 'yyy'
    assert path_out.is_file()
    assert path_out.read_text() == 'XyyyX'


def test_objectrewriter_in_to_out_no_out_path(temp_file_creator):
    """Object Rewriter in place edit with path object."""
    representer = ArbRepresenter()
    rewriter = filesystem.ObjectRewriter(get_arb_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(path_in, None)

    assert representer.load_payload == 'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == 'XyyyX'
    assert representer.dump_call_count == 1

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call("opening temp file for writing..."),
        call(f"moving temp file to: {path_in}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'


def test_objectrewriter_in_to_out_no_out_path_binary(temp_file_creator):
    """Object Rewriter in place edit with path object in binary mode."""
    representer = ArbRepresenter(True)
    rewriter = filesystem.ObjectRewriter(get_binary_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    rewriter.in_to_out(path_in, None)

    assert representer.load_payload == b'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == b'XyyyX'
    assert representer.dump_call_count == 1

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'


def test_objectrewriter_in_to_out_no_out_str(temp_file_creator):
    """Object Rewriter in place edit with str object."""
    representer = ArbRepresenter()
    rewriter = filesystem.ObjectRewriter(get_arb_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(str(path_in), None)

    assert representer.load_payload == 'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == 'XyyyX'
    assert representer.dump_call_count == 1

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call("opening temp file for writing..."),
        call(f"moving temp file to: {path_in}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'


def test_objectrewriter_in_to_out_same_path(temp_file_creator):
    """Object Rewriter in place edit with path object."""
    representer = ArbRepresenter()
    rewriter = filesystem.ObjectRewriter(get_arb_formatted,
                                         representer)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    str_out = os.path.relpath(path_in)

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(path_in, str_out)

    assert representer.load_payload == 'yyy'
    assert representer.load_call_count == 1
    assert representer.dump_payload == 'XyyyX'
    assert representer.dump_call_count == 1

    assert mock_logger_debug.mock_calls == [
        call("in path and out path are the same file. writing to temp "
             "file and then replacing in path with the temp file."),
        call(f"opening source file: {path_in}"),
        call("opening temp file for writing..."),
        call(f"moving temp file to: {path_in}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'
# endregion ObjectRewriter

# region StreamRewriter


def test_streamrewriter_in_to_out_path(temp_dir, temp_file_creator):
    """Stream Rewriter writes single in to out path object."""
    rewriter = filesystem.StreamRewriter(get_arb_formatted_iter)

    path_in = temp_file_creator()
    path_in.write_text('yyy')
    path_out = temp_dir.joinpath('sub', 'filename')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(path_in, path_out)

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call(f"opening destination file for writing: {path_out}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'yyy'
    assert path_out.is_file()
    assert path_out.read_text() == 'XyyyX'


def test_streamrewriter_in_to_out_str(temp_dir, temp_file_creator):
    """Stream Rewriter writes single in to out path object."""
    rewriter = filesystem.StreamRewriter(get_arb_formatted_iter)

    path_in = temp_file_creator()
    path_in.write_text('yyy')
    path_out = temp_dir.joinpath('sub', 'filename')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(str(path_in), str(path_out))

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call(f"opening destination file for writing: {path_out}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'yyy'
    assert path_out.is_file()
    assert path_out.read_text() == 'XyyyX'


def test_streamrewriter_in_to_out_no_out_path(temp_file_creator):
    """Stream Rewriter in place edit with path object."""
    rewriter = filesystem.StreamRewriter(get_arb_formatted_iter)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(path_in, None)

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call("opening temp file for writing..."),
        call(f"moving temp file to: {path_in}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'


def test_streamrewriter_in_to_out_no_out_str(temp_file_creator):
    """Stream Rewriter in place edit with str object."""
    rewriter = filesystem.StreamRewriter(get_arb_formatted_iter)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(str(path_in), None)

    assert mock_logger_debug.mock_calls == [
        call(f"opening source file: {path_in}"),
        call("opening temp file for writing..."),
        call(f"moving temp file to: {path_in}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'


def test_streamrewriter_in_to_out_same_path(temp_file_creator):
    """Stream Rewriter in place edit with same path."""
    rewriter = filesystem.StreamRewriter(get_arb_formatted_iter)

    path_in = temp_file_creator()
    path_in.write_text('yyy')

    str_out = os.path.relpath(path_in)

    with patch_logger(
            'pypyr.utils.filesystem', logging.DEBUG
    ) as mock_logger_debug:
        rewriter.in_to_out(str(path_in), str_out)

    assert mock_logger_debug.mock_calls == [
        call("in path and out path are the same file. writing to temp "
             "file and then replacing in path with the temp file."),
        call(f"opening source file: {path_in}"),
        call("opening temp file for writing..."),
        call(f"moving temp file to: {path_in}")]

    assert path_in.is_file()
    assert path_in.read_text() == 'XyyyX'

# endregion StreamRewriter

# region ObjectRepresenter


def test_jsonrepresenter_loads():
    """Json Representer loads."""
    representer = filesystem.JsonRepresenter()
    with open('./tests/testfiles/test.json', representer.read_mode) as file:
        obj = representer.load(file)

    assert obj
    assert obj['key1'] == 'value1'
    assert obj['key2'] == 'value2'
    assert obj['key3'] == 'value3'


def test_jsonrepresenter_dumps(temp_file_creator):
    """Json Representer dumps."""
    payload = {
        'key1': 'value1',
        'key2': 'value2',
        'key3': [0, 1, 2]
    }
    representer = filesystem.JsonRepresenter()
    file_path = temp_file_creator()
    with open(file_path, representer.write_mode) as file:
        representer.dump(file, payload)

    assert file_path.read_text() == ('{\n'
                                     '  "key1": "value1",\n'
                                     '  "key2": "value2",\n'
                                     '  "key3": [\n'
                                     '    0,\n'
                                     '    1,\n'
                                     '    2\n'
                                     '  ]\n'
                                     '}')


def test_yamlrepresenter_loads():
    """Yaml Representer loads."""
    representer = filesystem.YamlRepresenter()
    with open('./tests/testfiles/test.yaml',
              representer.read_mode,
              encoding='utf-8') as file:
        obj = representer.load(file)

    assert obj
    assert obj == {'key': 'value1 !£$%# *',
                   'key2': 'blah',
                   'key3': ['l1',
                            '!£$% *',
                            'l2',
                            [
                                'l31',
                                {'l32': ['l321', 'l322']}]
                            ]}


def test_yamlrepresenter_dumps(temp_file_creator):
    """Yaml Representer dumps."""
    payload = {
        'key1': 'value1',
        'key2': 'value2',
        'key3': [0, 1, 2]
    }
    representer = filesystem.YamlRepresenter()
    file_path = temp_file_creator()
    with open(file_path, representer.write_mode) as file:
        representer.dump(file, payload)

    assert file_path.read_text() == ('key1: value1\n'
                                     'key2: value2\n'
                                     'key3:\n'
                                     '  - 0\n'
                                     '  - 1\n'
                                     '  - 2\n')
# endregion ObjectRepresenter

# region ensure_dir


def test_ensure_dir(temp_dir):
    """Ensure dir creates parent path with string input."""
    path = temp_dir.joinpath('sub1', 'sub2')

    filesystem.ensure_dir(str(path))
    assert temp_dir.joinpath('sub1').is_dir()


def test_ensure_dir_path(temp_dir):
    """Ensure dir creates parent path with path input."""
    path = temp_dir.joinpath('sub1', 'sub2')

    filesystem.ensure_dir(path)
    assert temp_dir.joinpath('sub1').is_dir()

# endregion ensure_dir

# region get_glob


def assert_list_of_paths_equal(obj, other):
    """Cross platform compare of list of paths."""
    assert set([Path(p) for p in obj]) == set([Path(p) for p in other])


def test_get_glob_str():
    """Glob takes strings."""
    paths = filesystem.get_glob('./tests/testfiles/t???.txt')

    assert_list_of_paths_equal(paths, ['./tests/testfiles/test.txt'])


def test_get_glob_path():
    """Glob takes paths."""
    paths = filesystem.get_glob(Path('./tests/testfiles/t???.txt'))
    assert_list_of_paths_equal(paths, ['./tests/testfiles/test.txt'])


def test_get_glob_list():
    """Glob takes list of str and paths."""
    paths = filesystem.get_glob([Path('./tests/testfiles/t???.txt'),
                                 './tests/testfiles/a*.py'])
    assert_list_of_paths_equal(paths, ['tests/testfiles/test.txt',
                                       './tests/testfiles/arb.py'])


def test_get_glob_chain():
    """Multiple results from iterable input chains into single output."""
    paths = filesystem.get_glob([Path('./tests/testfiles/glob/a*.1'),
                                 './tests/testfiles/glob/a*2',
                                 './tests/testfiles/glob/a*.3'])
    assert_list_of_paths_equal(paths,
                               ['tests/testfiles/glob/arb.1.1',
                                'tests/testfiles/glob/arb.1',
                                './tests/testfiles/glob/arb.2',
                                './tests/testfiles/glob/arb.2.2',
                                './tests/testfiles/glob/arb.3'])


def test_get_glob_recurse():
    """Double asterix recurses. Also returns directories."""
    paths = filesystem.get_glob([Path('./tests/testfiles/glob/**/a*.1'),
                                 './tests/testfiles/glob/**/a*2',
                                 './tests/testfiles/glob/**/a*.3'])

    assert_list_of_paths_equal(paths,
                               ['tests/testfiles/glob/arb.1.1',
                                'tests/testfiles/glob/arb.1',
                                'tests/testfiles/glob/sub/asub.1',
                                'tests/testfiles/glob/sub/arbsub.1',
                                './tests/testfiles/glob/arb.2',
                                './tests/testfiles/glob/arb.2.2',
                                './tests/testfiles/glob/sub/arbsub.2',
                                './tests/testfiles/glob/arb.3',
                                './tests/testfiles/glob/sub/arbsub.3'])


def test_get_glob_wrong_type():
    """Wrong type raises."""
    with pytest.raises(TypeError) as err:
        filesystem.get_glob(1)

    assert str(err.value) == ("path should be string, path-like or a list. "
                              "Instead, it's a <class 'int'>")


def test_get_glob_path_doesnt_exist():
    """Non existent paths return nothing."""
    paths = filesystem.get_glob(['./arbX', './arb/**/x'])
    assert not paths
# endregion get_glob

# region is_same_file


def test_is_same_file():
    """Same file is same."""
    assert filesystem.is_same_file('./tests/testfiles/test.txt',
                                   './tests/testfiles/test.txt')


def test_is_same_file_path():
    """Same file is same with path input."""
    assert filesystem.is_same_file(Path('./tests/testfiles/test.txt'),
                                   Path('./tests/testfiles/test.txt'))


def test_is_same_file_relative_paths_resolve():
    """Same file with relative path obfuscation."""
    assert filesystem.is_same_file(
        './tests/testfiles/test.txt',
        './tests/testfiles/../../tests/testfiles/test.txt')


def test_is_same_file_dirs():
    """Directories aren't files."""
    assert not filesystem.is_same_file('./tests', './../tests/')


def test_is_same_file_one_exists_other_doesnt():
    """One file exists, the other doesn't."""
    assert not filesystem.is_same_file('./tests/testfiles/test.txt', './arb')


def test_is_same_file_paths_dont_exist():
    """Nonsense paths not same."""
    assert not filesystem.is_same_file('arb/arb', 'arb/arb')
    assert not filesystem.is_same_file('arb/arb1', 'arb/arb')
    assert not filesystem.is_same_file('arb/arb2', None)
    assert not filesystem.is_same_file(None, None)
# endregion is_same_file

# region move_file


def test_move_file_new(temp_dir, temp_file_creator):
    """Move file moves file if destination doesn't exist."""
    file1 = temp_file_creator()
    with open(file1, 'w') as f1:
        f1.write('arb file content')

    file2 = temp_dir.joinpath('file2')

    filesystem.move_file(file1, file2)
    assert not file1.is_file()
    assert file2.is_file()

    with open(file2) as f2:
        assert f2.read() == 'arb file content'


def test_move_file_new_str(temp_dir, temp_file_creator):
    """Move file moves file if destination doesn't exist, input str."""
    file1 = temp_file_creator()
    with open(file1, 'w') as f1:
        f1.write('arb file content')

    file2 = temp_dir.joinpath('file2')

    filesystem.move_file(str(file1), str(file2))
    assert not file1.is_file()
    assert file2.is_file()

    with open(file2) as f2:
        assert f2.read() == 'arb file content'


def test_move_file_overwrite(temp_file_creator):
    """Move file overwrites existing file."""
    file1 = temp_file_creator()
    with open(file1, 'w') as f1:
        f1.write('arb file content')

    file2 = temp_file_creator()

    filesystem.move_file(file1, file2)
    assert not file1.is_file()
    assert file2.is_file()

    with open(file2) as f2:
        assert f2.read() == 'arb file content'


def test_move_file_dest_parent_not_exist(temp_dir, temp_file_creator):
    """Move file fails if destination parent doesn't exist."""
    file1 = temp_file_creator()

    file2 = temp_dir.joinpath('blah', 'file2')

    with pytest.raises(FileNotFoundError) as err:
        filesystem.move_file(file1, file2)

    the_err = err.value
    assert file1 == Path(the_err.filename)
    assert file2 == Path(the_err.filename2)

    assert file1.is_file()
    assert not file2.is_file()
# endregion move_file

# region move_temp_file


def test_move_temp_file_new(temp_dir, temp_file_creator):
    """Move temp file moves file if destination doesn't exist."""
    file1 = temp_file_creator()
    with open(file1, 'w') as f1:
        f1.write('arb file content')

    file2 = temp_dir.joinpath('file2')

    filesystem.move_temp_file(file1, file2)
    assert not file1.is_file()
    assert file2.is_file()

    with open(file2) as f2:
        assert f2.read() == 'arb file content'


def test_move_temp_file_dest_parent_not_exist(temp_dir, temp_file_creator):
    """Move temp file fails if destination parent doesn't exist, wipes src."""
    file1 = temp_file_creator()

    file2 = temp_dir.joinpath('blah', 'file2')

    with pytest.raises(FileNotFoundError) as err:
        filesystem.move_temp_file(file1, file2)

    the_err = err.value
    assert file1 == Path(the_err.filename)
    assert file2 == Path(the_err.filename2)

    assert not file1.is_file()
    assert not file2.is_file()


@patch('os.remove', side_effect=ValueError('test remove failed'))
def test_move_temp_file_err_removing_src(mock_remove,
                                         temp_dir,
                                         temp_file_creator):
    """Move temp file fails if can't remove src on clean-up."""
    file1 = temp_file_creator()

    file2 = temp_dir.joinpath('blah', 'file2')

    with patch_logger(
            'pypyr.utils.filesystem', logging.ERROR
    ) as mock_logger_err:
        with pytest.raises(FileNotFoundError) as err:
            filesystem.move_temp_file(file1, file2)

    the_err = err.value
    assert file1 == Path(the_err.filename)
    assert file2 == Path(the_err.filename2)

    assert mock_logger_err.mock_calls == [
        call(f"error moving file {file1} to {file2}. {the_err}"),
        call(f'error removing temp file {file1}. test remove failed')]

    assert file1.is_file()
    assert not file2.is_file()


def test_move_temp_file_new_relative_paths(temp_dir, temp_file_creator):
    """Move temp file moves file if destination doesn't exist."""
    file1 = temp_file_creator()
    with open(file1, 'w') as f1:
        f1.write('arb file content')

    file2 = temp_dir.joinpath('file2')

    # work out relative paths to cwd
    rel_path1 = os.path.relpath(file1)
    rel_path2 = os.path.relpath(file2)

    filesystem.move_temp_file(rel_path1, rel_path2)
    assert not file1.is_file()
    assert file2.is_file()

    with open(file2) as f2:
        assert f2.read() == 'arb file content'

# endregion move__temp_file
